# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Optimization and DOE history post processing and analysis
"""
import os

# drop here optimizaiton / DOE history analysis
# to generate plots or quantitative criteria


# To be used for Jenkins which has no display, otherwise graphs export
# wont work
NO_DISPLAY = os.environ.get("NO_DISPLAY")
if NO_DISPLAY == "True":

    import matplotlib

    from gemseo import LOGGER

    LOGGER.warning("Use 'Agg' in matplotlib prevents from displaying windows")
    matplotlib.use("Agg")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A factory to execute post processings from their class name
***********************************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.core.factory import Factory
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.utils.py23_compat import string_types

standard_library.install_aliases()


from gemseo import LOGGER


class PostFactory(object):
    """Post processing factory to run optimization post processings
    Lists available post processings on the current configuration,
    executes them on demand.

    Works both from memory, from a ran optimization problem,
    and from disk, from a serialized optimization problem.
    """

    def __init__(self):
        """
        Initializes the factory: scans the directories to search for
        subclasses of OptPostProcessor.
        Searches in "GEMSEO_PATH" and gemseo.post
        """
        self.factory = Factory(OptPostProcessor, ("gemseo.post",))
        self.executed_post = []

    @property
    def posts(self):
        """Lists the available post processings

        :returns: the list of methods
        """
        return self.factory.classes

    def is_available(self, name):
        """Checks the availability of a post processing name

        :param name: the name of the post processing
        :returns: True if the post step is installed
        """
        return self.factory.is_available(name)

    def create(self, opt_problem, post_name):
        """Factory method to create a post processing subclass from post_name
        which is a class name

        :param opt_problem: the optimization problem on which to run
            the post processing
        :param post_name: the post processing name

        """
        return self.factory.create(post_name, opt_problem=opt_problem)

    def execute(self, opt_problem, post_name, **options):
        """Finds the appropriate library and executes
        the post processing on the problem

        :param opt_problem: the optimization problem on which to run
            the post procesing
        :param post_name: the post processing name
        """
        if isinstance(opt_problem, string_types):
            opt_problem = OptimizationProblem.import_hdf(opt_problem)
        post = self.create(opt_problem, post_name)
        post.execute(**options)
        self.executed_post.append(post)
        return post

    def list_generated_plots(self):
        """Lists the generated plot files"""
        plots = []
        for post in self.executed_post:
            plots.extend(post.output_files)
        return set(plots)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Box plots to quantify optimum robustness
****************************************
"""
from __future__ import absolute_import, division, unicode_literals

from math import sqrt

import matplotlib.pyplot as plt
import numpy.random as npr
from future import standard_library
from numpy import zeros

from gemseo.post.core.robustness_quantifier import RobustnessQuantifier
from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()
from gemseo import LOGGER


class Robustness(OptPostProcessor):
    """
    The **Robustness** post processing
    performs a quadratic approximation from an optimization history,
    and plot the results as cuts of the approximation
    computes the quadratic approximations of all the output functions,
    propagate analytically a normal distribution centered on the optimal
    design variable with a standard deviation which is a percentage
    of the mean passed in option (default: 1%)
    and plot the corresponding output boxplot.

    It is possible either to save the plot, to show the plot or both.
    """

    SR1_APPROX = "SR1"

    def _plot(
        self, save=True, show=False, stddev=0.01, file_path="boxplot", extension="pdf"
    ):
        """
        Builds the plot and saves it

        :param function: function name to build quadratic approximation
        :type function: bool
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param stddev: standard deviation of inputs as fraction of x bounds
        :type stddev: float
        :param share_y: if True, all Y axis are the same,
            useful to comare sensitivities
        :type share_y: bool
        :param extension: file extension
        :type extension: str
        """
        fig = self.__boxplot(stddev)
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )

    def __boxplot(self, stddev=0.01):
        """
        Plots the Hessian of the function

        :param stddev : standard deviation of inputs as fraction of x bounds
        """
        robustness = RobustnessQuantifier(self.database, "SR1")
        n_x = self.opt_problem.get_dimension()
        cov = zeros((n_x, n_x))
        upper_bounds = self.opt_problem.design_space.get_upper_bounds()
        lower_bounds = self.opt_problem.design_space.get_lower_bounds()
        bounds_range = upper_bounds - lower_bounds
        cov[list(range(n_x)), list(range(n_x))] = (stddev * (bounds_range)) ** 2

        data = []
        funcs_names = []
        for func in self.opt_problem.get_all_functions():
            func_name = func.name
            dim = func.dim
            for i in range(dim):
                b0_mat = zeros((n_x, n_x))
                robustness.compute_approximation(
                    funcname=func_name,
                    at_most_niter=int(1.5 * n_x),
                    func_index=i,
                    b0_mat=b0_mat,
                )
                x_ref = robustness.x_ref
                mean = robustness.compute_expected_value(x_ref, cov)
                var = robustness.compute_variance(x_ref, cov)
                if var > 0:  # Otherwise normal doesnt work
                    data.append(npr.normal(loc=mean, scale=sqrt(var), size=500))
                    legend = func_name
                    if dim > 1:
                        legend += "_" + str(i + 1)
                    funcs_names.append(legend)

        fig = plt.figure()
        fig.suptitle(
            "Box plot of the optimization functions "
            + "with normalized stddev "
            + str(stddev)
        )
        plt.boxplot(data, showfliers=False, labels=funcs_names)

        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Dimension reduction
===================

The :class:`.DimensionReduction` class implements the concept of dimension
reduction.

.. seealso::

   :mod:`~gemseo.mlearning.transform.dimension_reduction.pca`
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.mlearning.transform.transformer import Transformer

standard_library.install_aliases()


class DimensionReduction(Transformer):
    """ Dimension reduction. """

    def __init__(self, name="DimensionReduction", n_components=5, **parameters):
        """Constructor.

        :param str name: name of the scaler.
        :param int n_components: number of components. Default: 5.
        :param parameters: parameters for the dimension reduction algorithm.
        """
        super(DimensionReduction, self).__init__(
            name, n_components=n_components, **parameters
        )

    def fit(self, data):
        """Fit dimension reduction algorithm to data.

        :param ndarray data: data to be fitted.
        """
        raise NotImplementedError

    @property
    def n_components(self):
        """ Number of components """
        return self.parameters["n_components"]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable disciplines from Tedford and Martins (2010)
****************************************************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import array, atleast_2d, diag, dot
from numpy import mean as npmean
from numpy import ndarray
from numpy import sum as npsum
from past.utils import old_div

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.scalable.scalable_tm.variables import (
    OBJECTIVE_NAME,
    X_SHARED_NAME,
    get_constraint_name,
    get_coupling_name,
    get_x_local_name,
)

standard_library.install_aliases()

from gemseo import LOGGER


class TMSystem(MDODiscipline):

    r"""The system discipline from the scalable problem introduced by Tedford
    and Martins (2010) takes the  local design parameters
    :math:`x_1,x_2,\ldots,x_N` and the global design parameters :math:`z`
    as inputs, as well as the coupling variables :math:`y_1,y_2,\ldots,y_N`
    and returns the objective function value :math:`f(x,y(x,y))` to minimize as
    well as the inequality constraints ones
    :math:`c_1(y_1),c_2(y_2),\ldots,c_N(y_N)` which are expressed as:

    .. math::

       f(z,y) = |z|_2^2 + \sum_{i=1}^N |y_i|_2^2


    and:

    .. math::

       c_i(y_i) = 1- C_i^{-T}Iy_i

    """

    def __init__(self, c_constraint, default_inputs):
        """Constructor

        :param list(array) c_constraint: constraint coefficients
        :param dict default_inputs: default inputs
        """
        super(TMSystem, self).__init__("TM_System")
        self.input_grammar.initialize_from_base_dict(default_inputs)
        default_outputs = {}
        default_outputs[OBJECTIVE_NAME] = array([0.0])
        for index, value in enumerate(c_constraint):
            tmp = array([0.0] * len(value))
            default_outputs[get_constraint_name(index)] = tmp
        self.output_grammar.initialize_from_base_dict(default_outputs)
        self.default_inputs = default_inputs
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        self.coefficients = c_constraint
        self.n_disciplines = len(c_constraint)

    def _run(self):
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        coupling = [
            self.get_inputs_by_name(get_coupling_name(index))
            for index in range(self.n_disciplines)
        ]
        obj = npmean(x_shared ** 2)
        obj += npmean(
            [npmean(coupling[index] ** 2) for index in range(self.n_disciplines)]
        )
        data = {}
        data[OBJECTIVE_NAME] = array([obj])
        for index in range(self.n_disciplines):
            constraint = get_constraint_name(index)
            tmp = old_div(coupling[index], self.coefficients[index])
            data[constraint] = 1 - tmp
        self.store_local_data(**data)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the jacobian

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        coupling = [
            self.get_inputs_by_name(get_coupling_name(index))
            for index in range(self.n_disciplines)
        ]
        tmp = old_div(2 * x_shared, len(x_shared))
        self.jac[OBJECTIVE_NAME][X_SHARED_NAME] = atleast_2d(tmp)
        for index in range(self.n_disciplines):
            tmp = old_div(2 * coupling[index], len(coupling[index]))
            tmp /= self.n_disciplines
            self.jac[OBJECTIVE_NAME][get_coupling_name(index)] = atleast_2d(tmp)
            tmp = diag(old_div(-1.0, self.coefficients[index]))
            constraint = get_constraint_name(index)
            coupling_name = get_coupling_name(index)
            self.jac[constraint][coupling_name] = atleast_2d(tmp)


class TMDiscipline(MDODiscipline):

    r"""An elementary discipline from the scalable problem introduced by
    Tedford and Martins (2010) takes local design parameters :math:`x_i`
    and shared design parameters :math:`z` in input as well as coupling
    variables :math:`\left(y_i\right)_{1\leq j \leq N\atop j\neq i}`
    from :math:`N-1` elementary disciplines,
    and returns the coupling variables:

    .. math::

        y_i =\frac{\tilde{y}_i+C_{z,i}.1+C_{x_i}.1}{\sum_{j=1 \atop j
        \neq i}^NC_{y_j,i}.1+C_{z,i}.1+C_{x_i}.1} \in [0,1]^{n_{y_i}}

    where:

    .. math::

        \tilde{y}_i = - C_{z,i}.z - C_{x_i}.x_i +
        \sum_{j=1 \atop j \neq i}^N C_{y_j,i}.y_j
    """

    def __init__(self, index, c_shared, c_local, c_coupling, default_inputs):
        """Constructor

        :param int index: discipline index for naming.
        :param array c_shared: weights for the shared design parameters.
        :param array c_local: weights for the local design parameters.
        :param dict(array) c_coupling: weights for the coupling parameters.
        :param dict default_inputs: default inputs
        """
        self.index = index
        self.c_shared = c_shared
        self.c_local = c_local
        self.c_coupling = c_coupling
        self.n_disciplines = len(c_coupling)
        self._check_consistency(default_inputs)
        super(TMDiscipline, self).__init__(name="TM_Discipline_" + str(index))
        self.input_grammar.initialize_from_base_dict(default_inputs)
        value = array([0.0] * self.c_shared.shape[0])
        default_outputs = {get_coupling_name(index): value}
        self.output_grammar.initialize_from_base_dict(default_outputs)
        self.default_inputs = default_inputs
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _check_consistency(self, default_inputs):
        """Check consistency of default inputs.

        :param dict default_inputs: default inputs.
        """
        assert isinstance(self.c_shared, ndarray)
        assert len(self.c_shared.shape) == 2
        assert self.c_shared.shape[1] == len(default_inputs[X_SHARED_NAME])
        nout = self.c_shared.shape[0]
        assert isinstance(self.c_local, ndarray)
        assert len(self.c_local.shape) == 2
        assert self.c_local.shape[0] == nout
        xlocal = get_x_local_name(self.index)
        assert self.c_local.shape[1] == len(default_inputs[xlocal])
        assert isinstance(self.c_coupling, dict)
        for key, val in self.c_coupling.items():
            assert isinstance(val, ndarray)
            assert len(val.shape) == 2
            assert val.shape[0] == nout
            assert val.shape[1] == len(default_inputs[key])

    def _run(self):
        index = self.index
        x_shared = self.get_inputs_by_name(X_SHARED_NAME)
        x_local = self.get_inputs_by_name(get_x_local_name(index))
        coupling = {
            cpl_name: self.get_inputs_by_name(cpl_name)
            for cpl_name in list(self.c_coupling.keys())
        }
        data = -dot(self.c_shared, x_shared.reshape((-1, 1)))
        data -= dot(self.c_local, x_local.reshape((-1, 1)))
        cpl_sum = 0
        for cpl_name, cpl_value in self.c_coupling.items():
            data += dot(cpl_value, coupling[cpl_name].reshape((-1, 1)))
            cpl_sum += npsum(self.c_coupling[cpl_name], 1)
        data += npsum(self.c_shared, 1).reshape((-1, 1))
        data += npsum(self.c_local, 1).reshape((-1, 1))
        norm = cpl_sum.reshape((-1, 1))
        norm += npsum(self.c_shared, 1).reshape((-1, 1))
        norm += npsum(self.c_local, 1).reshape((-1, 1))
        data /= norm
        data = data.flatten()
        self.store_local_data(**{get_coupling_name(index): data})

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the jacobian

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        self._init_jacobian(inputs, outputs, with_zeros=True)
        other_indices = set(range(self.n_disciplines)) - set([self.index])
        other_indices = list(other_indices)
        cpl_sum = 0
        for cpl_value in self.c_coupling.values():
            cpl_sum += npsum(cpl_value, 1)
        norm = cpl_sum.reshape((-1, 1))
        norm += npsum(self.c_shared, 1).reshape((-1, 1))
        norm += npsum(self.c_local, 1).reshape((-1, 1))
        der = old_div(-self.c_local, norm)
        self.jac[get_coupling_name(self.index)][get_x_local_name(self.index)] = der
        der = old_div(-self.c_shared, norm)
        self.jac[get_coupling_name(self.index)][X_SHARED_NAME] = der
        for cpl_name, cpl_value in self.c_coupling.items():
            self.jac[get_coupling_name(self.index)][cpl_name] = old_div(cpl_value, norm)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A factory to instantiate formulation from their class names
***********************************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library

from gemseo.core.factory import Factory
from gemseo.core.formulation import MDOFormulation

standard_library.install_aliases()


from gemseo import LOGGER


class MDOFormulationsFactory(object):
    """MDO Formulations factory to create the formulation from a name
    or a class.
    """

    def __init__(self):
        """
        Initializes the factory: scans the directories to search for
        subclasses of MDOFormulation.
        Searches in "GEMSEO_PATH" and gemseo.formulations
        """
        # Defines the benchmark problems to be imported
        self.factory = Factory(MDOFormulation, ("gemseo.formulations",))

    def create(
        self, formulation_name, disciplines, objective_name, design_space, **options
    ):
        """
        Create a formulation from its name

        :param formulation_name: the formulation name,
            the class name of the formulation in gemseo.formulations
        :param disciplines: list of disciplines
        :param objective_name: the objective function name
        :param design_space: the design space
        :param options: options for creation of the formulation
        """
        return self.factory.create(
            formulation_name,
            disciplines=disciplines,
            design_space=design_space,
            objective_name=objective_name,
            **options
        )

    @property
    def formulations(self):
        """
        Lists the available classes

        :returns : the list of classes names
        """
        return self.factory.classes

    def is_available(self, formulation_name):
        """
        Checks the availability of a formulation

        :param name : formulation_name of the formulation
        :returns: True if the formulation is available
        """
        return self.factory.is_available(formulation_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A Bi level formulation
**********************
"""
from __future__ import absolute_import, division, unicode_literals

from builtins import str, super

from future import standard_library

from gemseo.core.chain import MDOChain, MDOParallelChain
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.formulation import MDOFormulation
from gemseo.core.function import MDOFunction
from gemseo.core.mdo_scenario import MDOScenarioAdapter
from gemseo.mda.mda_factory import MDAFactory

standard_library.install_aliases()


from gemseo import LOGGER


class BiLevel(MDOFormulation):
    """
    A bi-level formulation draws an optimization architecture
    that involves multiple optimization problems to be solved to
    obtain the solution of the MDO problem.

    Here, at each iteration on the global design variables,
    the bi-level MDO formulation implementation performs a first
    MDA to compute the coupling variables, then disciplinary
    optimizations on the local design variables in parallel and then,
    a second MDA to update the coupling variables.
    """

    def __init__(
        self,
        disciplines,
        objective_name,
        design_space,
        maximize_objective=False,
        mda_name="MDAChain",
        parallel_scenarios=False,
        multithread_scenarios=True,
        apply_cstr_tosub_scenarios=True,
        apply_cstr_to_system=True,
        reset_x0_before_opt=False,
        **mda_options
    ):
        """
        Constructor, initializes the objective functions and constraints

        :param disciplines: the disciplines list.
        :type disciplines: list(MDODiscipline)
        :param objective_name: the objective function data name.
        :type objective_name: str
        :param design_space: the design space.
        :type design_space: DesignSpace
        :param maximize_objective: if True, the objective function
            is maximized, by default, a minimization is performed.
        :type maximize_objective: bool
        :param mda_name: class name of the MDA to be used.
        :type mda_name: str
        :param parallel_scenarios: if True, the sub scenarios are run
            in parallel.
        :type parallel_scenarios: bool
        :param multithread_scenarios: if True and parallel_scenarios=True,
             the sub scenarios are run in parallel using multi-threading,
             if False and parallel_scenarios=True, multi-processing is used.
        :type multithread_scenarios: bool
        :param apply_cstr_tosub_scenarios: if True, the add_constraint
            method adds the constraint to the optimization problem
            of the sub-scenario capable of computing the constraint.
        :type apply_cstr_tosub_scenarios: bool
        :param apply_cstr_to_system: if True, the add_constraint
            method adds the constraint to the optimization problem
            of the system scenario.
        :type apply_cstr_to_system: bool
        :param reset_x0_before_opt: if True, restart the sub optimizations
            from the initial guesses, otherwise warm start them
        :type reset_x0_before_opt: bool
        :param mda_options: options passed to the MDA at construction
        """
        super(BiLevel, self).__init__(
            disciplines,
            objective_name,
            design_space,
            maximize_objective=maximize_objective,
        )
        self._shared_dv = list(design_space.variables_names)
        self.mda1 = None
        self.mda2 = None
        self.reset_x0_before_opt = reset_x0_before_opt
        self.scenario_adapters = []
        self.chain = None
        self._mda_factory = MDAFactory()
        self._apply_cstr_to_system = apply_cstr_to_system
        self._apply_cstr_tosub_scenarios = apply_cstr_tosub_scenarios
        self.__parallel_scenarios = parallel_scenarios
        self._multithread_scenarios = multithread_scenarios
        self.couplstr = MDOCouplingStructure(self.get_sub_disciplines())

        # Create MDA
        self._build_mdas(mda_name, **mda_options)

        # Create MDOChain : MDA1 -> sub scenarios -> MDA2
        self._build_chain()

        # Cleanup design space
        self._update_design_space()

        # Builds the objective function on top of the chain
        self._build_objective_from_disc(self._objective_name)

    def _build_scenario_adapters(
        self,
        output_functions=False,
        pass_nonshared_var=False,
        adapter_class=MDOScenarioAdapter,
        **adapter_options
    ):
        """Builds the MDOScenarioAdapter required for each sub scenario
        This is used to build the self.chain.

        :param output_functions: if True then the optimization functions are
            outputs of the adapter
        :type output_functions: bool
        :param pass_nonshared_var: If True, the non-shared design variables
            are inputs of the scenarios adapters
        :type pass_nonshared_var: bool
        :param adapter_class: class of the adapters
        :type adapter_class: MDOScenarioAdapter
        :param adapter_options: options for the adapters initialization
        :type adapter_options: dict
        """
        adapters = []
        # coupled sub-disciplines

        couplings = self.couplstr.strong_couplings()
        mda2_inpts = self._get_mda2_inputs()

        shared_dv = set(self._shared_dv)
        for scenario in self.get_sub_scenarios():

            # Get the I/O names of the sub-scenario top-level disciplines
            top_disc = scenario.formulation.get_top_level_disc()
            top_inputs = [
                inpt for disc in top_disc for inpt in disc.get_input_data_names()
            ]
            top_outputs = [
                outpt for disc in top_disc for outpt in disc.get_output_data_names()
            ]

            # All couplings of the scenarios are taken from the MDA
            sc_allins = list(
                set(top_inputs) & set(couplings)
                |
                # Add shared variables from system scenario driver
                set(top_inputs) & shared_dv
            )
            if pass_nonshared_var:
                nonshared_var = scenario.design_space.variables_names
                sc_allins = list(set(sc_allins) | set(top_inputs) & set(nonshared_var))
            # Output couplings of scenario are given to MDA for speedup
            if output_functions:
                opt_problem = scenario.formulation.opt_problem
                sc_outvars = opt_problem.objective.outvars
                sc_constraints = opt_problem.get_constraints_names()
                sc_out_coupl = sc_outvars + sc_constraints
            else:
                sc_out_coupl = list(set(top_outputs) & set(couplings + mda2_inpts))

            # Add private variables from disciplinary scenario design space
            sc_allouts = sc_out_coupl + scenario.design_space.variables_names

            adapter = adapter_class(scenario, sc_allins, sc_allouts, **adapter_options)
            adapters.append(adapter)
        return adapters

    def _get_mda2_inputs(self):
        """Return the list of MDA2 inputs."""
        return []

    @classmethod
    def get_sub_options_grammar(cls, **options):
        """
        When some options of the formulation depend on higher level
        options, a sub option schema may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options grammar
        """
        main_mda = options.get("mda_name")
        if main_mda is None:
            raise ValueError(
                "'mda_name' option is required \n"
                + "to deduce the sub options of BiLevel !"
            )
        factory = MDAFactory().factory
        return factory.get_options_grammar(main_mda)

    @classmethod
    def get_default_sub_options_values(cls, **options):
        """
        When some options of the formulation depend on higher level
        options, a sub option defaults may be specified here, mainly for
        use in the API

        :param options: options dict required to deduce the sub options grammar
        :returns: None, or the sub options defaults
        """
        main_mda = options.get("mda_name")
        if main_mda is None:
            raise ValueError(
                "'mda_name' option is required \n"
                + "to deduce the sub options of BiLevel !"
            )
        factory = MDAFactory().factory
        return factory.get_default_options_values(main_mda)

    def _build_mdas(self, mda_name, **mda_options):
        """Builds the chain : MDA -> MDOScenarios -> MDA
        on top of which all functions are built.
        """
        disc_mda1 = self.couplstr.strongly_coupled_disciplines()
        if len(disc_mda1) > 0:
            self.mda1 = self._mda_factory.create(mda_name, disc_mda1, **mda_options)
            self.mda1.warm_start = True
        else:
            LOGGER.warning(
                "No strongly coupled disciplines detected, "
                + " MDA1 is deactivated in the BiLevel formulation."
            )

        disc_mda2 = self.get_sub_disciplines()
        self.mda2 = self._mda_factory.create(mda_name, disc_mda2, **mda_options)

        self.mda2.warm_start = False

    def _build_chain_dis_sub_opts(self):
        """
        Inits the chain of disciplines and the list of sub scenarios
        """
        chain_dis = []
        if self.mda1 is not None:
            chain_dis = [self.mda1]
        sub_opts = self.scenario_adapters
        return chain_dis, sub_opts

    def _build_chain(self):
        """Builds the chain : MDA -> MDOScenarios -> MDA
        on top of which all functions are built.
        """
        # Build the scenario adapters to be chained with MDAs
        adapter_opt = {"reset_x0_before_opt": self.reset_x0_before_opt}
        self.scenario_adapters = self._build_scenario_adapters(**adapter_opt)
        chain_dis, sub_opts = self._build_chain_dis_sub_opts()

        if self.__parallel_scenarios:
            use_threading = self._multithread_scenarios

            par_chain = MDOParallelChain(sub_opts, use_threading=use_threading)

            chain_dis += [par_chain, self.mda2]
        else:
            # Chain MDA -> scenarios exec -> MDA
            chain_dis += sub_opts + [self.mda2]

        self.chain = MDOChain(chain_dis, name="bilevel_chain")

        if not self.reset_x0_before_opt and self.mda1 is not None:
            run_mda1_orig = self.mda1._run

            def _run_mda():
                """Redefine mda1 execution to warm start the chain
                with previous x_local opt

                :param input_data: Default value = None)
                """
                # TODO : Define a pre run method to be overloaded in MDA maybe
                # Or use observers at the system driver level to pass the local
                # vars
                for scenario in self.get_sub_scenarios():
                    x_loc_d = scenario.design_space.get_current_x_dict()
                    for indata, x_loc in x_loc_d.items():
                        if self.mda1.is_input_existing(indata):
                            if x_loc is not None:
                                self.mda1.local_data[indata] = x_loc
                return run_mda1_orig()

            self.mda1._run = _run_mda

    def _update_design_space(self):
        """Update the design space by removing the coupling variables"""
        self._set_defaultinputs_from_ds()
        self._remove_sub_scenario_dv_from_ds()
        self._remove_couplings_from_ds()
        self._remove_unused_variables()

    def _remove_couplings_from_ds(self):
        """Removes the coupling variables from the design space"""
        if hasattr(self.mda2, "strong_couplings"):
            # Otherwise, the MDA2 may be a user provided MDA
            # Which manages the couplings internally
            couplings = self.mda2.strong_couplings
            design_space = self.opt_problem.design_space
            for coupling in couplings:
                if coupling in design_space.variables_names:
                    design_space.remove_variable(coupling)

    def get_top_level_disc(self):
        """ Overriden method from MDOFormulation base class """
        return [self.chain]

    def get_expected_workflow(self):
        """Overriden method from MDOFormulation base class
        delegated to chain object"""
        return self.chain.get_expected_workflow()

    def get_expected_dataflow(self):
        """Overriden method from MDOFormulation base class
        delegated to chain object"""
        return self.chain.get_expected_dataflow()

    def add_constraint(
        self,
        output_name,
        constraint_type=MDOFunction.TYPE_EQ,
        constraint_name=None,
        value=None,
        positive=False,
    ):
        """
        Add a contraint to the formulation

        :param output_name: param constraint_type:
            (Default value = MDOFunction.TYPE_EQ)
        :param constraint_name: Default value = None)
        :param value: Default value = None)
        :param positive: Default value = False)
        :param constraint_type:  (Default value = MDOFunction.TYPE_EQ)

        """
        if self._apply_cstr_to_system:
            super(BiLevel, self).add_constraint(
                output_name, constraint_type, constraint_name, value, positive
            )
        if self._apply_cstr_tosub_scenarios:
            added = False
            outputs_list = self._check_add_cstr_input(output_name, constraint_type)
            for scen in self.get_sub_scenarios():
                if self._scenario_computes_outputs(scen, outputs_list):
                    scen.add_constraint(
                        outputs_list, constraint_type, constraint_name, value, positive
                    )
                    added = True
            if not added:
                raise ValueError(
                    "No sub scenario has an output named "
                    + str(output_name)
                    + " cannot create such a constraint."
                )

    @staticmethod
    def _scenario_computes_outputs(scenario, output_names):
        """Returns True if the top level disciplines compute
        the outputs named output_names

        :param output_names: name of the variable names to check
        :param scenario: the scenario to be tested
        """
        for disc in scenario.formulation.get_top_level_disc():
            if disc.is_all_outputs_existing(output_names):
                return True
        return False



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
ZvsXY plot
==========

A :class:`.ZvsXY` plot represents the variable :math:`z` with respect to
:math:`x` and :math:`y` as a surface plot, based on a set of points
:points :math:`\{x_i,y_i,z_i\}_{1\leq i \leq n}`. This interpolation is
relies on the Delaunay triangulation of :math:`\{x_i,y_i\}_{1\leq i \leq n}`
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from future import standard_library

from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


class ZvsXY(DatasetPlot):
    """ Plot surface z versus x,y. """

    def _plot(
        self,
        x,
        y,
        z,
        x_comp=0,
        y_comp=0,
        z_comp=0,
        colormap="Blues",
        add_points=False,
        color="blue",
    ):
        """Surface.

        :param x: name of the variable on the x-axis
        :type x: str
        :param y: name of the variable on the y-axis
        :type z: str
        :param z: name of the variable on the z-axis
        :type z: str
        :param x_comp: x component. Default: 0.
        :type x_comp: int
        :param y_comp: y component. Default: 0.
        :type y_comp: int
        :param z_comp: z component. Default: 0.
        :type z_comp: int
        :param colormap: matplotlib colormap. Default: 'Blues'.
        :type colormap: str
        :param add_points: display points over the surface plot.
            Default: False.
        :type add_points: bool
        :param color: point color. Default: 'blue'.
        :type color: str
        """
        x_data = self.dataset[x][x][:, x_comp]
        y_data = self.dataset[y][y][:, y_comp]
        z_data = self.dataset[z][z][:, z_comp]

        fig = plt.figure()
        axes = fig.add_subplot(1, 1, 1)
        triang = mtri.Triangulation(x_data, y_data)
        tcf = axes.tricontourf(triang, z_data, cmap=colormap)
        if add_points:
            axes.scatter(x_data, y_data, color=color)
        if self.dataset.sizes[x] == 1:
            axes.set_xlabel(x)
        else:
            axes.set_xlabel(x + "(" + str(x_comp) + ")")
        if self.dataset.sizes[y] == 1:
            axes.set_ylabel(y)
        else:
            axes.set_ylabel(y + "(" + str(y_comp) + ")")
        if self.dataset.sizes[z] == 1:
            axes.set_title(z)
        else:
            axes.set_title(z + "(" + str(z_comp) + ")")
        fig.colorbar(tcf)
        fig = plt.gcf()
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Classification model factory
============================

This module contains a factory to instantiate a :class:`.MLClassificationAlgo`
from its class name. The class can be internal to |g| or located in an
external module whose path is provided to the constructor. It also provides a
list of available classification models and allows you to test if a
classification model type is available.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.core.factory import Factory
from gemseo.mlearning.classification.classification import MLClassificationAlgo
from gemseo.mlearning.core.factory import MLAlgoFactory

standard_library.install_aliases()


from gemseo import LOGGER


class ClassificationModelFactory(MLAlgoFactory):
    """This factory instantiates a :class:`.MLRegressionAlgo`
    from its class name. The class can be internal to |g| or located in an
    external module whose path is provided to the constructor.
    """

    def __init__(self):
        """Initializes the factory: scans the directories to search for
        subclasses of :class:`.MLClassificationAlgo`. Searches in "GEMSEO_PATH"
        and gemseo.mlearning.classification.
        """
        super(ClassificationModelFactory, self).__init__()
        self.factory = Factory(
            MLClassificationAlgo, ("gemseo.mlearning.classification",)
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
The notions of distribution and composed distributions
======================================================

Overview
--------

The abstract :class:`.Distribution` class implements the concept of
probability distribution. It is enriched by concrete classes
such as :class:`.OTDistribution` and :class:`.SPDistribution`.

Similarly, the abstract :class:`.ComposedDistribution` class implements
the concept of composed probability distribution from a list
of probability distributions. It inherits from :class:`.Distribution`.

Construction
------------

The :class:`.Distribution` of a given uncertain variable is built
from a recognized distribution name,
a variable dimension, a set of parameters
and optionally a standard representation of these parameters.

The :class:`.ComposedDistribution` of a list of given uncertain variables
is built from a list of :class:`.Distribution` objects
implementing the probability distributions of these variables
and from a copula name.

.. note::

   A copula is a mathematical function used to define the dependence
   between random variables from their cumulative density functions.
   `See more <https://en.wikipedia.org/wiki/Copula_(probability_theory)>`_.

Capabilities
------------

From a :class:`.Distribution`, we can easily get statistics,
such as :attr:`.Distribution.mean`,
:attr:`.Distribution.standard_deviation`,
numerical :attr:`.Distribution.range` and
mathematical :attr:`.Distribution.support`,
or plot the cumulative and probability density functions,
either for a given marginal (:meth:`.Distribution.plot`)
or for all marginals (:meth:`.Distribution.plot_all`).

We can also compute the cumulative density function
(:meth:`.Distribution.cdf`)
for the different marginals of the random variable,
as well as the inverse cumulative density function
(:meth:`.Distribution.inverse_cdf`).

Lastly, we can get realizations of the random variable
by means of the :meth:`.Distribution.get_sample` method.

.. note::

   As :class:`.ComposedDistribution` inherits from :class:`.Distribution`,
   it has the same capabilities.
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
from future import standard_library
from numpy import arange, array, concatenate
from past.utils import old_div

standard_library.install_aliases()

from gemseo import LOGGER


class Distribution(object):
    """ Distribution. """

    MATHEMATICAL_LOWER_BOUND = "math_l_b"
    MATHEMATICAL_UPPER_BOUND = "math_u_b"
    NUMERICAL_LOWER_BOUND = "num_l_b"
    NUMERICAL_UPPER_BOUND = "num_u_b"
    MU = "mu"
    SIGMA = "sigma"
    LOWER = "lower"
    UPPER = "upper"
    MODE = "mode"
    RATE = "rate"
    LOC = "loc"

    def __init__(
        self, variable, distribution, parameters, dimension, standard_parameters=None
    ):
        """Constructor

        :param str variable: variable name.
        :param str distribution: distribution name.
        :param parameters: distribution parameters.
        :type parameters: tuple or dict
        :param int dimension: variable dimension.
        :param dict standard_parameters: standard parameters.
        """
        self.math_lower_bound = None
        self.math_upper_bound = None
        self.num_lower_bound = None
        self.num_upper_bound = None
        self.distribution = None
        self.marginals = None
        self.dimension = dimension
        self.variable_name = variable
        self.distribution_name = distribution
        self.transformation = variable
        self.parameters = parameters
        if standard_parameters is None:
            self.standard_parameters = self.parameters
        else:
            self.standard_parameters = standard_parameters
        LOGGER.info("Define the random variable: %s ", variable)
        LOGGER.info("|_ Distribution: %s ", str(self))
        LOGGER.info("|_ Dimension: %s ", str(dimension))

    def __str__(self):
        """String representation of the object.

        :return: string representation.
        :rtype: str
        """
        label = self.distribution_name + "("
        if isinstance(self.standard_parameters, dict):
            param = iter(self.standard_parameters.items())
            param = sorted([name + "=" + str(val) for name, val in param])
            label += ", ".join(param)
        else:
            param = [str(option) for option in self.standard_parameters]
            label += ", ".join(param)
        label += ")"
        return label

    def get_sample(self, n_samples=1):
        """Get several samples.

        :param int n_samples: number of samples.
        :return: samples
        :rtype: array
        """
        raise NotImplementedError

    def cdf(self, vector):
        """Evaluate the cumulative density function of the random variable
        marginals for a given instance.

        :param array vector: instance of the random variable.
        :return: cdf values
        :rtype: array
        """
        raise NotImplementedError

    def inverse_cdf(self, vector):
        """Evaluate the inverse of the cumulative density function of the
        random variable marginals for a given unit vector .

        :param array vector: vector of values comprised between 0 and 1 with
            same dimension as the random variable.
        :return: inverse cdf values
        :rtype: array
        """
        raise NotImplementedError

    @property
    def mean(self):
        """Get the mean of the random variable.

        :return: mean of the random variable.
        :rtype: array
        """
        raise NotImplementedError

    @property
    def standard_deviation(self):
        """Get the standard deviation of the random variable.

        :return: standard deviation of the random variable.
        :rtype: array
        """
        raise NotImplementedError

    @property
    def range(self):
        """Get the numerical range for the different components.

        :return: numerical range.
        :rtype: list(array)
        """
        value = [
            array([l_b, u_b])
            for l_b, u_b in zip(self.num_lower_bound, self.num_upper_bound)
        ]
        return value

    @property
    def support(self):
        """Get the mathematical support for the different components.

        :return: mathematical support.
        :rtype: list(array)
        """
        value = [
            array([l_b, u_b])
            for l_b, u_b in zip(self.math_lower_bound, self.math_upper_bound)
        ]
        return value

    def plot_all(self, show=True, save=False, prefix=None):
        """Plot the probability density function and the cumulative density
        function of the random variable.

        :param str prefix: If not None, start the filename with a prefix.
            Default: None.
        """
        for index in range(self.dimension):
            self.plot(index, show, save, prefix)

    def plot(self, index=0, show=True, save=False, prefix=None):
        """Plot the probability density function and the cumulative density
        function of the random variable.

        :param str prefix: If not None, start the filename with a prefix.
            Default: None.
        """
        variable_name = self.variable_name + "(" + str(index) + ")"
        l_b = self.num_lower_bound[index]
        u_b = self.num_upper_bound[index]
        xvals = arange(l_b, u_b, old_div((u_b - l_b), 100))
        y1vals = [self._pdf(index)(xval) for xval in xvals]
        ax1 = plt.subplot(121)
        ax1.plot(xvals, y1vals)
        ax1.set_xlabel(variable_name)
        ax1.set_title("PDF")
        y2vals = [self._cdf(index)(xval) for xval in xvals]
        ax2 = plt.subplot(122)
        ax2.plot(xvals, y2vals)
        ax2.set_xlabel(variable_name)
        ax2.set_title("CDF")
        fname = "distribution_" + self.variable_name + "_" + str(index)
        if show:
            plt.show()
        if save:
            if prefix is not None:
                fname = prefix + "_" + fname
            plt.savefig(fname + ".pdf")

    def _pdf(self, index):
        """Get the probability density function of a marginal.

        :param int index: marginal index.
        :return: probability density function
        :rtype: function
        """

        def pdf(point):
            """Probability Density Function (PDF).

            :param float point: point value.
            :return: PDF value.
            :rtype: float
            """
            raise NotImplementedError

        return pdf

    def _cdf(self, index):
        """Get the cumulative density function of a marginal.

        :param int index: marginal index.
        :return: cumulative density function
        :rtype: function
        """

        def cdf(level):
            """Cumulative Density Function (CDF).

            :param float point: point value.
            :return: CDF value.
            :rtype: float
            """
            raise NotImplementedError

        return cdf


class ComposedDistribution(Distribution):
    """ Composed Distribution. """

    INDEPENDENT_COPULA = "independent_copula"
    AVAILABLE_COPULA = [INDEPENDENT_COPULA]
    COMPOSED = "Composed"

    def __init__(self, distributions, copula=INDEPENDENT_COPULA):
        """Constructor.

        :param list(Distribution) distributions: list of distributions.
        :param str copula: copula name. Default: INDEPENDENT_COPULA.
        """
        dimension = sum([distribution.dimension for distribution in distributions])
        self._marginal_variables = [
            distribution.variable_name for distribution in distributions
        ]
        variable = "_".join(self._marginal_variables)
        super(ComposedDistribution, self).__init__(
            variable, self.COMPOSED, (copula,), dimension
        )
        self.marginals = distributions
        LOGGER.info("|_ Marginals:")
        for distribution in distributions:
            LOGGER.info(
                "   - %s(%s): %s",
                distribution.variable_name,
                distribution.dimension,
                distribution,
            )

    def _set_bounds(self, distributions):
        """Set mathematical and numerical bounds (= support and range).

        :param list(Distribution) distributions: list of distributions.
        """
        self.math_lower_bound = array([])
        self.math_upper_bound = array([])
        self.num_lower_bound = array([])
        self.num_upper_bound = array([])
        for dist in distributions:
            self.math_lower_bound = concatenate(
                (self.math_lower_bound, dist.math_lower_bound)
            )
            self.num_lower_bound = concatenate(
                (self.num_lower_bound, dist.num_lower_bound)
            )
            self.math_upper_bound = concatenate(
                (self.math_upper_bound, dist.math_upper_bound)
            )
            self.num_upper_bound = concatenate(
                (self.num_upper_bound, dist.num_upper_bound)
            )

    @property
    def mean(self):
        """Get the mean of the random variable.

        :return: mean of the random variable.
        :rtype: array
        """
        mean = [marginal.mean for marginal in self.marginals]
        return array(mean).flatten()

    @property
    def standard_deviation(self):
        """Get the standard deviation of the random variable.

        :return: standard deviation of the random variable.
        :rtype: array
        """
        std = [marginal.standard_deviation for marginal in self.marginals]
        return array(std).flatten()

    def get_sample(self, n_samples=1):
        """Get sample.

        :param int n_samples: number of samples.
        :return: samples
        :rtype: list(array)
        """
        sample = self.marginals[0].get_sample(n_samples)
        for marginal in self.marginals[1:]:
            sample = concatenate((sample, marginal.get_sample(n_samples)), axis=1)
        return sample



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Data scaler
===========

The :class:`.Scaler` class implements the default scaling method applying to
some parameter :math:`z`:

.. math::

    \\bar{z} := \\text{offset} + \\text{coefficient}\\times z

where :math:`\\bar{z}` is the scaled version of z. This scaling method is a
linear transformation parameterized by an offset and a coefficient.

In this default scaling method, the offset is equal to 0 and the coefficient is
equal to 1. Consequently, the scaling operation is the identity:
:math:`\\bar{z}=z`. This method has to be overloaded.

.. seealso::

   :mod:`~gemseo.mlearning.transform.scaler.min_max_scaler`
   :mod:`~gemseo.mlearning.transform.scaler.standard_scaler`
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import diag, eye, ndarray

from gemseo.mlearning.transform.transformer import Transformer

standard_library.install_aliases()


class Scaler(Transformer):
    """ Data scaler. """

    def __init__(self, name="Scaler", offset=0.0, coefficient=1.0):
        """Constructor.

        :param str name: name of the scaler.
        :param float offset: offset of the linear transformation.
            Default: 0.
        :param float coefficient: coefficient of the linear transformation.
            Default: 1.
        """
        super(Scaler, self).__init__(name, offset=offset, coefficient=coefficient)

    @property
    def offset(self):
        """ Offset. """
        return self.parameters["offset"]

    @property
    def coefficient(self):
        """ Coefficient. """
        return self.parameters["coefficient"]

    @offset.setter
    def offset(self, value):
        """ Set offset. """
        self.parameters["offset"] = value

    @coefficient.setter
    def coefficient(self, value):
        """ Set coefficient. """
        self.parameters["coefficient"] = value

    def fit(self, data):
        """Fit scaler to data. Offset and coefficient terms are already
        defined in the constructor.

        :param ndarray data: data to be fitted.
        """
        return

    def transform(self, data):
        """Scale data using the offset and coefficient terms.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        scaled_data = self.offset + self.coefficient * data
        return scaled_data

    def inverse_transform(self, data):
        """Unscale data using the offset and coefficient terms.

        :param ndarray data: data to be inverse transformed.
        :return: inverse transformed data.
        :rtype: ndarray
        """
        unscaled_data = (data - self.offset) / self.coefficient
        return unscaled_data

    def compute_jacobian(self, data):
        """Compute Jacobian of the scaler transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        if not isinstance(self.coefficient, ndarray):
            jac = self.coefficient * eye(data.shape[-1])
        else:
            jac = diag(self.coefficient)
        return jac

    def compute_jacobian_inverse(self, data):
        """Compute Jacobian of the scaler inverse_transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        if not isinstance(self.coefficient, ndarray):
            jac_inv = 1 / self.coefficient * eye(data.shape[-1])
        else:
            jac_inv = diag(1 / self.coefficient)
        return jac_inv



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Random forest classification model
==================================

The random forest classification model uses averaging methods on an ensemble
of decision trees.

Dependence
----------
The classifier relies on the RandomForestClassifier class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.ensemble.RandomForestClassifier.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import stack
from sklearn.ensemble import RandomForestClassifier as SKLRandForest

from gemseo.mlearning.classification.classification import MLClassificationAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class RandomForestClassifier(MLClassificationAlgo):
    """ Random forest classification algorithm. """

    LIBRARY = "scikit-learn"
    ABBR = "RandomForestClassifier"

    def __init__(
        self,
        data,
        transformer=None,
        input_names=None,
        output_names=None,
        n_estimators=100,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param n_estimators: number of trees in the forest.
        :type n_estimators: int
        :param parameters: other keyword arguments for sklearn rand. forest.
        """
        super(RandomForestClassifier, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            n_estimators=n_estimators,
            **parameters
        )
        self.algo = SKLRandForest(n_estimators=n_estimators, **parameters)

    def _fit(self, input_data, output_data):
        """Fit the classification model.

        :param ndarray input_data: input data (2D).
        :param ndarray(int) output_data: output data.
        """
        if output_data.shape[1] == 1:
            output_data = output_data.ravel()
        self.algo.fit(input_data, output_data)

    def _predict(self, input_data):
        """Predict output data from input data.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: output data (n_samples, n_outputs).
        :rtype: ndarray(int)
        """
        output_data = self.algo.predict(input_data).astype(int)
        if len(output_data.shape) == 1:
            output_data = output_data[:, None]
        return output_data

    def _predict_proba_soft(self, input_data):
        """Predict probability of belonging to each class.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: probabilities of belonging to each class
            (n_samples, n_outputs, n_classes). For a given sample and output
            variable, the sum of probabilities is one.
        :rtype: ndarray
        """
        probas = self.algo.predict_proba(input_data)
        if len(probas[0].shape) == 1:
            probas = probas[..., None]
        else:
            probas = stack(probas, axis=-1)
        return probas



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Python2 and Python3 compatibility layer
"""

import sys
from shutil import rmtree
from tempfile import mkdtemp

from numpy import array
from six import string_types

PY2 = sys.version_info.major == 2
PY3 = not PY2
PY36 = sys.version_info.major == 3 and sys.version_info.minor >= 6

if PY3:
    from builtins import int as _long

    string_dtype = "bytes"
    from builtins import range as xrange
    from inspect import getfullargspec as _getfullargspec

    from fastjsonschema import compile as compile_schema
    from fastjsonschema.exceptions import JsonSchemaException

    def next(iterator):
        return iterator.__next__()


else:
    from builtins import long as _long
    from builtins import xrange

    string_dtype = "string"
    from builtins import next
    from inspect import getargspec as _getfullargspec

    from gemseo.third_party.fastjsonschema import compile as compile_schema
    from gemseo.third_party.fastjsonschema.exceptions import JsonSchemaException

if PY36:
    OrderedDict = dict
else:
    from collections import OrderedDict


def is_py2():
    """Check if the version of Python is 2.

    :return: True if Python 2.
    :rtype: bool
    """
    return PY2


def long(data):
    """
    Return a long from data, performs casting
    """
    return _long(data)


def string_array(data):
    """
    Creates a numpy array of strings from data
    the dtype is adjusted (bytes in py3, string in py2)
    """
    return array(data, dtype=string_dtype)


if PY2:

    class TemporaryDirectory(object):
        """Create and return a temporary directory.  This has the same
        behavior as mkdtemp but can be used as a context manager.  For
        example:
            with TemporaryDirectory() as tmpdir:
                ...
        Upon exiting the context, the directory and everything contained
        in it are removed.

        Backported from python3 to python2 by Francois Gallard,
        with big simplification that may lead to some issues...
        """

        def __init__(self, suffix="", prefix="tmp", dir=None):
            self.name = None
            self.name = mkdtemp(suffix, prefix, dir)

        def __repr__(self):
            return "<{} {!r}>".format(self.__class__.__name__, self.name)

        def __enter__(self):
            return self.name

        def __exit__(self, exc, value, tb):
            self.cleanup()

        def __del__(self):
            self.cleanup()

        def cleanup(self):
            """ Delete the entire directory tree. """
            rmtree(self.name)


else:
    from tempfile import TemporaryDirectory

if PY2:
    getargspec = _getfullargspec

    def to_unicode_list(iterable):
        """ Convert a list of strings to a list of unicode strings."""
        return [s.decode("utf-8") for s in iterable]


else:

    def getargspec(func):
        """ Get arguments specifications. """
        return _getfullargspec(func)[:4]

    def to_unicode_list(iterable):
        """ Convert a list of strings to a list of unicode strings."""
        return iterable



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Remi Lafage
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Generate XDSMjs input json file
*******************************
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import webbrowser
from builtins import open, str
from json import dumps
from multiprocessing import RLock
from os import getcwd
from os.path import abspath, basename, join, splitext
from tempfile import mkdtemp

from future import standard_library

from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import (
    AtomicExecSequence,
    LoopExecSequence,
    ParallelExecSequence,
    SerialExecSequence,
)
from gemseo.core.mdo_scenario import MDOScenarioAdapter
from gemseo.core.monitoring import Monitoring
from gemseo.mda.mda import MDA
from gemseo.utils.locks import synchronized
from gemseo.utils.show_utils import generate_xdsm_html

from .xdsm_tikz import xdsm_dict2tex

standard_library.install_aliases()


from gemseo import LOGGER

OPT_NAME = OPT_ID = "Opt"
USER_NAME = USER_ID = "_U_"


class XDSMizer(object):
    """Class to build the XDSM diagram of a scenario

    Generates input json for XDSMjs javascript library
    https://github.com/OneraHub/XDSMjs

    See :
    Martins, Joaquim RRA, and Andrew B. Lambe.
    "Multidisciplinary design optimization: a survey of architectures."
    AIAA journal 51.9 (2013): 2049-2075.
    """

    def __init__(self, scenario, hashref="root", level=0, expected_workflow=None):
        """
        Constructor

        :param scenario: the MDO scenario to be represented as a XDSM diagram
        :param hashref: key used in the final XDSM json structure to reference
                        {nodes, edges, workflow, optpb} data structure.
                        Default to 'root'
        :param level: level number corresponding to scenario depth.
                      Root scenario is level 0.
        """
        self.scenario = scenario
        self.level = level
        self.hashref = hashref
        self.lock = RLock()
        self._monitor = None
        self.outdir = "."
        self.outfilename = "xdsm.json"
        self.to_hashref = {}
        self.to_id = {}  # dictionary to map AtomicExecSequence to XDSM id
        self.initialize(expected_workflow)
        self.print_statuses = False  # Prints the statuses in the console
        self.latex_output = False

    def initialize(self, workflow=None):
        """
        Initialize from a given workflow or use self.scenario' one,
        create sub XDSM diagram accordingly.

        :param workflow: composite execution sequence
        """
        self.sub_xdsmizers = []
        # Find disciplines from workflow structure
        if workflow:
            self.workflow = workflow
        else:
            self.workflow = self.scenario.get_expected_workflow()
        self.atoms = XDSMizer._get_single_level_atoms(self.workflow)

        self.to_hashref = {}
        level = self.level + 1
        num = 1
        for atom in self.atoms:
            if atom.discipline.is_scenario():
                if atom.discipline == self.scenario:
                    self.to_hashref[atom] = "root"
                    self.root_atom = atom
                else:  # sub-scenario
                    self.to_hashref[atom] = "scn-" + str(level) + "-" + str(num)
                    sub_workflow = XDSMizer._find_sub_workflow(self.workflow, atom)
                    self.sub_xdsmizers.append(
                        XDSMizer(
                            atom.discipline, self.to_hashref[atom], level, sub_workflow
                        )
                    )
                    num += 1

    def monitor(
        self,
        outdir=".",
        outfilename="xdsm.json",
        print_statuses=False,
        latex_output=False,
    ):
        """Monitors |g| discipline execution by generating XDSM json file on
        discipline status update.

        :param str outdir: the directory where XDSM json file is generated
        :param str outfilename: the file name of the generated XDSM json file
            (default: xdsm.json)
        :param bool print_statuses: print the statuses in the console
            at each update (default: False)
        :param bool latex_output: generate tikz, tex and pdf output
            (default: False)
        """
        self._monitor = Monitoring(self.scenario)
        self._monitor.add_observer(self)
        # have to reinitialize with monitored workflow
        self.initialize(self._monitor.workflow)
        self.outdir = outdir
        self.outfilename = outfilename
        self.print_statuses = print_statuses
        self.latex_output = latex_output

    def update(self, atom):  # pylint: disable=unused-argument
        """Callback function that generate new XDSM regarding the given
        atom status update
        :param atom: discipline which status has been updated
        """
        self.run(
            outdir=self.outdir,
            outfilename=self.outfilename,
            latex_output=self.latex_output,
        )
        if self.print_statuses:
            LOGGER.info(str(self._monitor))

    def run(
        self,
        outdir=None,
        latex_output=False,
        outfilename="xdsm.html",
        html_output=True,
        json_output=False,
        open_browser=False,
    ):
        """Generates a XDSM diagram from the process.
        By default, a self contined HTML file is generated, that can be
        viewed in a browser.


        :param outdir: the directory where XDSM json file is generated.
            if None, current working dir is used.
            If open_browser is True and outdir is None, generates a
            temporary directory to store the file
        :param outfilename: the file name of the generated XDSM json file
            (default: xdsm.json).
        :param latex_output: produces .tex, .tikz and .tex files if True
            If ``outdir`` is not set the XDSM json is printed
            on the standard output.
        :param open_browser: if True, opens the web browser with the XDSM
        :param html_output: if True, outputs a self contained HTML file
        :param json_output: if True, outputs a JSON file for XDSMjs

        :returns: XDSM json either in a file when ``outdir`` is set,
                  ouput on the console otherwise.
        """
        xdsm = self.xdsmize()
        xdsm_json = dumps(xdsm, indent=2, ensure_ascii=False)
        base = basename(outfilename)
        outfile_basename = splitext(base)[0]

        no_html_loc = False
        if outdir is None:
            outdir = getcwd()
            no_html_loc = True

        if json_output:
            with open(join(outdir, outfile_basename + ".json"), "w") as out:
                out.write(xdsm_json)
        if latex_output:
            xdsm_dict2tex(xdsm, outdir, outfile_basename)

        if html_output or open_browser:
            if no_html_loc:
                outdir = mkdtemp(suffix="", prefix="tmp", dir=None)
            out_file_path = join(outdir, outfile_basename + ".html")
            LOGGER.info("Generating HTML XDSM file in : %s", out_file_path)
            generate_xdsm_html(xdsm, out_file_path)
            if open_browser:
                url = "file://" + abspath(out_file_path)
                webbrowser.open(url, new=2)  # open in new tab
            return out_file_path

        return xdsm

    def get_all_sub_xdsmizers(self):
        """Retrieves all sub xdsmizers corresponding to sub Scenario objects

        :returns: the array of XDSMizer objects
        """
        result = []
        for sub in self.sub_xdsmizers:
            result.append(sub)
            result.extend(sub.get_all_sub_xdsmizers())
        return result

    @synchronized
    def xdsmize(self, algoname="Optimizer"):
        """Builds the Python data structure to be used to generate JSON
        format compatible with XDSMjs viewer.

        :param algoname: Default value = "Optimizer")
        :returns: the Python object containing relevant information
                    for XDSM representation
        """
        nodes = self._create_nodes(algoname)
        edges = self._create_edges()
        workflow = self._create_workflow()
        optpb = str(self.scenario.formulation.opt_problem)

        if self.level == 0:
            res = {
                self.hashref: {
                    "nodes": nodes,
                    "edges": edges,
                    "workflow": workflow,
                    "optpb": optpb,
                }
            }
            for sub_xdsmizer in self.get_all_sub_xdsmizers():
                res[sub_xdsmizer.hashref] = sub_xdsmizer.xdsmize()
            return res
        return {"nodes": nodes, "edges": edges, "workflow": workflow, "optpb": optpb}

    def _create_nodes(self, algoname):  # pylint: disable=too-many-branches
        """Manages XDSM diagram nodes creation from optimization
        algorithm and disciplines

        :param algoname: name of the algorithm
        """
        nodes = []
        self.to_id = {}

        statuses = self.workflow.get_state_dict()

        # Optimization
        self.to_id[self.root_atom] = OPT_ID
        opt_node = {"id": OPT_ID, "name": algoname, "type": "optimization"}
        if statuses[self.root_atom.uuid]:
            opt_node["status"] = statuses[self.root_atom.uuid]

        nodes.append(opt_node)

        # Disciplines
        for atom_id, atom in enumerate(
            self.atoms
        ):  # pylint: disable=too-many-nested-blocks
            # if a node already created from an atom with same discipline
            # at one level just reference the same node
            for ref_atom in self.to_id:
                if atom.discipline == ref_atom.discipline:
                    self.to_id[atom] = self.to_id[ref_atom]

                    if (
                        atom.status
                        and atom.parent.status is MDODiscipline.STATUS_RUNNING
                    ):

                        node = None
                        for a_node in nodes:
                            if a_node["id"] == self.to_id[atom]:
                                node = a_node
                                break

                        if not node:
                            # TODO: add specific exception?
                            raise "Node " + self.to_id[
                                atom
                            ] + " not found in " + nodes  # pragma: no cover

                        node["status"] = atom.status

                    break

            if atom in self.to_id:
                continue

            self.to_id[atom] = "Dis" + str(atom_id)
            node = {"id": self.to_id[atom], "name": atom.discipline.name}

            # node type
            if isinstance(atom.discipline, MDA):
                node["type"] = "mda"
            elif atom.discipline.is_scenario():
                node["type"] = "mdo"
                node["subxdsm"] = self.to_hashref[atom]
                node["name"] = atom.discipline.name + "_" + self.to_hashref[atom]
            else:
                node["type"] = "analysis"

            if statuses[atom.uuid]:
                node["status"] = statuses[atom.uuid]

            nodes.append(node)

        return nodes

    def _create_edges(self):
        """Manage XDSM edges creation from scenario dataflow."""
        edges = []
        # convenient method to factorize code for creating and appending edges

        def add_edge(from_edge, to_edge, varnames):
            """Adds an edge

            :param from_edge: param to_edge:
            :param varnames:
            :param to_edge:

            """
            edge = {"from": from_edge, "to": to_edge, "name": ", ".join(varnames)}
            edges.append(edge)

        # For User to/from optimization
        opt_pb = self.scenario.formulation.opt_problem

        # fct names such as -y4
        functions_names = opt_pb.get_all_functions_names()

        # output variables used by the fonction (eg y4)
        fct_varnames = [f.outvars for f in opt_pb.get_all_functions()]
        function_varnames = []
        for fvars in fct_varnames:
            function_varnames.extend(fvars)

        to_user = functions_names
        to_opt = self.scenario.get_optim_variables_names()

        add_edge(USER_ID, OPT_ID, [x + "^(0)" for x in to_opt])
        add_edge(OPT_ID, USER_ID, [x + "^*" for x in to_user])

        # Disciplines to/from optimization
        for atom in self.atoms:
            if atom is not self.root_atom:
                varnames = set(atom.discipline.get_input_data_names()) & set(
                    self.scenario.get_optim_variables_names()
                )
                if varnames:
                    add_edge(OPT_ID, self.to_id[atom], varnames)

                varnames = set(atom.discipline.get_output_data_names()) & set(
                    function_varnames
                )
                # print set(disc.get_output_data_names()), set(functions_names)
                if varnames:
                    add_edge(self.to_id[atom], OPT_ID, varnames)

        # Disciplines to User/Optimization (from User is already handled at
        # optimizer level)
        disc_to_opt = function_varnames
        for atom in self.atoms:
            if atom is not self.root_atom:
                # special case MDA : skipped
                if isinstance(atom.discipline, MDA):
                    continue
                out_to_user = [
                    o
                    for o in atom.discipline.get_output_data_names()
                    if o not in disc_to_opt
                ]
                out_to_opt = [
                    o
                    for o in atom.discipline.get_output_data_names()
                    if o in disc_to_opt
                ]
                if out_to_user:
                    add_edge(self.to_id[atom], USER_ID, [x + "^*" for x in out_to_user])
                if out_to_opt:
                    add_edge(self.to_id[atom], OPT_ID, out_to_opt)

        # Disciplines to/from disciplines
        for coupling in self.scenario.get_expected_dataflow():
            (disc1, disc2, varnames) = coupling
            add_edge(
                self.to_id[self._find_atom(disc1)],
                self.to_id[self._find_atom(disc2)],
                varnames,
            )

        return edges

    @staticmethod
    def _get_single_level_atoms(workflow):
        """
        Retrieves the list of atoms of the given workflow without
        looking into loop execution sequences coming from Scenario.
        Thus retrieves the atoms for a one level XDSM diagram.

        :param workflow: execution sequence
        :returns: a list of atoms
        """
        atoms = []
        for seq in workflow.sequence_list:
            if isinstance(seq, LoopExecSequence):
                atoms.append(seq.atom_controller)
                if not seq.atom_controller.discipline.is_scenario():
                    atoms += XDSMizer._get_single_level_atoms(seq.iteration_sequence)
            elif isinstance(seq, AtomicExecSequence):
                atoms.append(seq)
            else:
                atoms += XDSMizer._get_single_level_atoms(seq)
        return atoms

    def _find_atom(self, discipline):
        """
        Find atom corresponding to the given discipline.
        Raise exception if not found

        :param disicpline: an MDODiscipline
        :returns: the corresponding atom
        """
        atom = None
        if isinstance(discipline, MDOScenarioAdapter):
            atom = self._find_atom(discipline.scenario)
        else:
            for atom_i in self.atoms:
                if discipline == atom_i.discipline:
                    atom = atom_i
        if atom is None:
            disciplines = [a.discipline for a in self.atoms]
            raise ValueError(
                "Discipline {} " "not found in {}".format(discipline, disciplines)
            )
        return atom

    @staticmethod
    def _find_sub_workflow(workflow, atom_controller):
        """
        Find loop execution sequence sub_workflow with the given
        atom as controller within the given workflow

        :param atom_controller: the AtomicExecSequence object that controls
        the LoopExecutionSequence object to find
        :returns: the subworkflow (LoopExecutionSequence), None if not found
        """
        sub_workflow = None
        for seq in workflow.sequence_list:
            if isinstance(seq, LoopExecSequence):
                if seq.atom_controller.uuid == atom_controller.uuid:
                    sub_workflow = seq
                    return sub_workflow

                sub_workflow = sub_workflow or XDSMizer._find_sub_workflow(
                    seq.iteration_sequence, atom_controller
                )
            elif not isinstance(seq, AtomicExecSequence):
                sub_workflow = sub_workflow or XDSMizer._find_sub_workflow(
                    seq, atom_controller
                )

        return sub_workflow

    def _create_workflow(self):
        """Manage XDSM workflow creation from formulation workflow"""
        workflow = [USER_ID, expand(self.workflow, self.to_id)]
        return workflow


def expand(wks, to_id):
    """Expands workflow structure as an ids structure using to_id mapping.
    The expansion preserve the structure while replacing the object by its id
    in all case except when a tuple is encountered as cdr then the expansion
    transforms loop[A, (B,C)] in [idA, {'parallel': [idB, idC]}]
    :param wks: the workflow structure
    :param to_id: the mapping dict from object to id
    :returns: the ids structure valid to be used as XDSM json chains
    """
    if isinstance(wks, SerialExecSequence):
        res = []
        for seq in wks.sequence_list:
            res += expand(seq, to_id)
        ids = res
    elif isinstance(wks, ParallelExecSequence):
        res = []
        for seq in wks.sequence_list:
            res += expand(seq, to_id)
        ids = [{"parallel": res}]
    elif isinstance(wks, LoopExecSequence):
        if (
            wks.atom_controller.discipline.is_scenario()
            and to_id[wks.atom_controller] != OPT_ID
        ):
            # sub-scnario consider only the controller
            ids = [to_id[wks.atom_controller]]
        else:
            ids = [to_id[wks.atom_controller], expand(wks.iteration_sequence, to_id)]
    elif isinstance(wks, AtomicExecSequence):
        ids = [to_id[wks]]
    else:
        raise Exception("Bad execution sequence : found " + str(wks))
    return ids



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Machine learning algorithm baseclass
====================================

Machine learning is the art of building models from data,
the latter being samples of properties of interest
that can sometimes be sorted by group, such as inputs, outputs, categories, ...

In the absence of such groups, the data can be analyzed through a study of
commonalities, leading to plausible clusters. This is referred to as
clustering, a branch of unsupervised learning dedicated to the detection
of patterns in unlabeled data.

.. seealso::

   :mod:`~gemseo.mlearning.core.unsupervised`,
   :mod:`~gemseo.mlearning.cluster.cluster`

When data can be separated into at least two categories by a human, supervised
learning can start with classification whose purpose is to model the relation
between these categories and the properties of interest. Once trained,
a classification model can predict the category corresponding to new
property values.

.. seealso::

   :mod:`~gemseo.mlearning.core.supervised`,
   :mod:`~gemseo.mlearning.classification.classification`

When the distinction between inputs and outputs can be made among the data
properties, another branch of supervised learning can be called: regression
modeling. Once trained, a regression model can predict the outputs
corresponding to new inputs values.

.. seealso::

   :mod:`~gemseo.mlearning.core.supervised`,
   :mod:`~gemseo.mlearning.regression.regression`

The quality of a machine learning algorithm can be measured
using a :class:`.MLQualityMeasure` either with respect to the learning dataset
or to a test dataset or using resampling methods, such as K-folds or
leave-one-out cross-validation techniques. The challenge is to avoid
over-learning the learning data leading to a loss of generality.
We often want to build models that are not too dataset-dependent. For that,
we want to maximize both a learning quality and a generalization quality.
In unsupervised learning, a quality measure can represent the robustness of
clusters definition while in supervised learning, a quality measure can be
interpreted as an error, whether it is a misclassification in the case of the
classification algorithms or a prediction one in the case of the regression
algorithms. This quality can often be improved by building machine learning
models from standardized data in such a way that the data properties have the
same order of magnitude.


.. seealso::

   :mod:`~gemseo.mlearning.qual_measure.quality_measure`,
   :mod:`~gemseo.mlearning.transform.transformer`

Lastly, a machine learning algorithm often depends on hyperparameters to
carefully tune in order to maximize the generalization power of the model.

.. seealso::

   :mod:`~gemseo.mlearning.core.calibration`

"""
from __future__ import absolute_import, division, unicode_literals

import pickle
import re
from os import makedirs
from os.path import exists, join

from future import standard_library

standard_library.install_aliases()


class MLAlgo(object):
    """The :class:`.MLAlgo` abstract class implements the concept of machine
    learning algorithm. Such a model is built from a training dataset,
    data transformation options and parameters. This abstract class defines the
    :meth:`.MLAlgo.learn`, :meth:`.MLAlgo.save` methods and the boolean
    property, :attr:`!MLAlgo.is_trained`. It also offers a string
    representation for end user.
    Inheriting classes should overload the :meth:`.MLAlgo.learn`,
    :meth:`!MLAlgo._save_algo` and :meth:`!MLAlgo._load_algo` methods.
    """

    LIBRARY = None
    ABBR = "MLAlgo"
    FILENAME = "ml_algo.pkl"

    def __init__(self, data, transformer=None, **parameters):
        """Constructor.

        :param Dataset data: learning dataset
        :param transformer: transformation strategy for data groups. If None,
            do not transform data. The dictionary keys are the groups to
            transform. The dictionary items are the transformers, either
            referenced by their names as strings, or provided directly.
            Default: None.
        :type transformer: dict(str) or dict(Transformer)
        :param parameters: algorithm parameters
        """
        self.learning_set = data
        self.parameters = parameters
        self.transformer = transformer or {}
        self.algo = None
        self._trained = False

    class DataFormatters(object):
        """ Decorators for internal MLAlgo methods. """

    @property
    def is_trained(self):
        """Check if algorithm is trained.
        :return: bool
        """
        return self._trained

    def learn(self, samples=None):
        """Train machine learning algorithm on learning set, possibly filtered
        using the given parameters.
        :param list(int) samples: indices of training samples.
        """
        raise NotImplementedError

    def __str__(self):
        """ String representation for end user. """
        string = self.__class__.__name__ + "("
        parameters = [key + "=" + str(val) for key, val in self.parameters.items()]
        string += ", ".join(parameters)
        string += ")\n"
        if self.LIBRARY is not None:
            string += "| based on the " + self.LIBRARY + " library\n"
        string += "| built from " + str(self.learning_set.length)
        string += " learning samples"
        return string

    def save(self, directory=None, path=".", save_learning_set=False):
        """Save the machine learning algorithm.

        :param str directory: directory name
        :param str path: path name
        :return: location of saved file
        :rtype: str
        """
        if not save_learning_set:
            self.learning_set.data = {}
            self.learning_set.length = 0
        splitted_class_name = re.findall("[A-Z][a-z]*", self.__class__.__name__)
        splitted_class_name = [word.lower() for word in splitted_class_name]
        algo_name = "_".join(splitted_class_name)
        algo_name += "_" + self.learning_set.name
        directory = directory or algo_name
        directory = join(path, directory)
        if not exists(directory):
            makedirs(directory)

        filename = join(directory, self.FILENAME)
        objects = self._get_objects_to_save()
        with open(filename, "wb") as handle:
            pickle.dump(objects, handle)

        self._save_algo(directory)

        return directory

    def _save_algo(self, directory):
        """Save external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        filename = join(directory, "algo.pkl")
        with open(filename, "wb") as handle:
            pickle.dump(self.algo, handle)

    def load_algo(self, directory):
        """Load external machine learning algorithm.

        :param str directory: algorithm directory.
        """

        filename = join(directory, "algo.pkl")
        with open(filename, "rb") as handle:
            algo = pickle.load(handle)
        self.algo = algo

    def _get_objects_to_save(self):
        """Get objects to save.

        :return: objects to save.
        :rtype: dict
        """
        objects = {
            "data": self.learning_set,
            "transformer": self.transformer,
            "parameters": self.parameters,
            "algo_name": self.__class__.__name__,
            "_trained": self._trained,
        }
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Damien Guenot
#                 Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Testing tools."""

import functools
from os import name as os_name

import pytest

from gemseo import LOGGER

IS_NT = os_name == "nt"

__skip_mp = pytest.mark.skipif(
    IS_NT, reason="Multiprocessing not available under Windows"
)


def __skip_nt(test_function):
    """
    Decorator used to skip a test under windows

    :param test_function: the function that is being tested and decorated
    :returns: the result of the wrapped function, if called, else None
    """
    if IS_NT:
        LOGGER.info("Skipping test %s", test_function.__name__)

        @functools.wraps(test_function)
        def wrapper(*args, **kwargs):  # pylint: disable=unused-argument
            pass

    else:

        @functools.wraps(test_function)
        def wrapper(*args, **kwargs):
            return test_function(*args, **kwargs)

    return wrapper


def skip_under_windows(test_function):
    """
    Decorator used to skip a test under windows
    Works both for unittests runs and pytest runs

    :param test_function: the function that is being tested and decorated
    :returns: the result of the wrapped function, if called, else None
    """
    return __skip_mp(__skip_nt(test_function))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Sellar MDO problem's Design Space
*********************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import super

import numpy as np
from future import standard_library

from gemseo.algos.design_space import DesignSpace

standard_library.install_aliases()


class SellarDesignSpace(DesignSpace):

    """**SellarDesignSpace** creates the :class:`.DesignSpace` of the
    Sellar problem
    whose :class:`.MDODiscipline` are :class:`.Sellar1`, :class:`.Sellar2`
    and :class:`.SellarSystem`.

    - x_local belongs to [0., 10.]
    - x_shared_0 belongs to [-10., 10.]
    - x_shared_1 belongs to [0., 10.]
    - y_0 belongs to [-100., 100.]
    - y_1 belongs to [-100., 100.]
    """

    def __init__(self, dtype="complex128"):
        """The constructor creates a blank :class:`.DesignSpace`
        to which it adds all design variables."""
        super(SellarDesignSpace, self).__init__()

        # construct a dictionary with initial solution
        x_local = np.array([1.0], dtype=dtype)
        x_shared = np.array([4.0, 3.0], dtype=dtype)
        y_0 = np.array([1.0], dtype=dtype)
        y_1 = np.array([1.0], dtype=dtype)

        # design variables
        self.add_variable("x_local", 1, l_b=0.0, u_b=10.0, value=x_local)  # x
        self.add_variable(
            "x_shared", 2, l_b=(-10, 0.0), u_b=(10.0, 10.0), value=x_shared
        )  # z

        # target coupling variables
        self.add_variable("y_0", 1, l_b=-100.0, u_b=100.0, value=y_0)
        self.add_variable("y_1", 1, l_b=-100.0, u_b=100.0, value=y_1)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Quality measure
===============

The :mod:`~gemseo.mlearning.qual_measure.quality_measure` module
implements the concept of quality measures for machine learning algorithms.

This concept is implemented through the :class:`.MLQualityMeasure` class.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

standard_library.install_aliases()


class MLQualityMeasure(object):
    """ Quality measure for machine learning. """

    LEARN = "learn"
    TEST = "test"
    LOO = "loo"
    KFOLDS = "kfolds"
    BOOTSTRAP = "bootstrap"

    def __init__(self, algo):
        """Constructor.

        :param MLAlgo algo: machine learning algorithm.
        """
        self.algo = algo

    def evaluate(self, method=LEARN, **options):
        """Evaluate quality measure.

        :param str method: method to estimate the quality measure.
        :param options: options of the estimation method (e.g. 'test_data' for
            the 'test' method, 'n_replicates' for the boostrap one, ...)
        :return: quality measure value.
        """
        if method == self.LEARN:
            evalutation = self.evaluate_learn(**options)
        elif method == self.TEST:
            evalutation = self.evaluate_test(**options)
        elif method == self.LOO:
            evalutation = self.evaluate_loo(**options)
        elif method == self.KFOLDS:
            evalutation = self.evaluate_kfolds(**options)
        elif method == self.BOOTSTRAP:
            evalutation = self.evaluate_bootstrap(**options)
        return evalutation

    def evaluate_learn(self, multioutput=True):
        """Evaluate quality measure using the learning dataset.

        :param bool multioutput: if True, return the quality measure for each
            output component. Otherwise, average these measures. Default: True.
        :return: quality measure value.
        """
        raise NotImplementedError

    def evaluate_test(self, test_data, multioutput=True):
        """Evaluate quality measure using a test dataset.

        :param Dataset test_data: test data.
        :param bool multioutput: if True, return the quality measure for each
            output component. Otherwise, average these measures. Default: True.
        :return: quality measure value.
        """
        raise NotImplementedError

    def evaluate_loo(self, multioutput=True):
        """Evaluate quality measure using the leave-one-out technique.

        :param bool multioutput: if True, return the quality measure for each
            component. Otherwise, average these measures. Default: True.
        :return: quality measure value.
        """
        n_samples = self.algo.learning_set.n_samples
        return self.evaluate_kfolds(n_folds=n_samples, multioutput=multioutput)

    def evaluate_kfolds(self, n_folds=5, multioutput=True):
        """Evaluate quality measure using the k-folds technique.

        :param int n_folds: number of folds. Default: 5.
        :param bool multioutput: if True, return the quality measure for each
            component. Otherwise, average these measures. Default: True.
        :return: quality measure value.
        """
        raise NotImplementedError

    def evaluate_bootstrap(self, n_replicates=100, multioutput=True):
        """Evaluate quality measure using the bootstrap technique.

        :param int n_replicates: number of bootstrap replicates. Default: 100.
        :param bool multioutput: if True, return the quality measure for each
            component. Otherwise, average these measures. Default: True.
        :return: quality measure value.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# pyxdsm is a python library for generating publication quality PDF XDSM
# diagrams.This library is a thin wrapper that uses the TIKZ library and
# LaTeX to build the PDFs. Source: https://github.com/mdolab/pyXDSM
#
# OpenMDAO Open Source License:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os

import numpy as np


tikzpicture_template = r"""
%%% Preamble Requirements %%%
% \usepackage{{geometry}}
% \usepackage{{amsfonts}}
% \usepackage{{amsmath}}
% \usepackage{{amssymb}}
% \usepackage{{sfmath}}
% \usepackage{{tikz}}

% \usetikzlibrary{{arrows,chains,positioning,scopes,shapes.geometric,shapes.misc,shadows}}

%%% End Preamble Requirements %%%

\input{{{diagram_styles_path}}}
\begin{{tikzpicture}}

\matrix[MatrixSetup]{{
{nodes}}};

% XDSM process chains
{process}

\begin{{pgfonlayer}}{{data}}
\path
{edges}
\end{{pgfonlayer}}

\end{{tikzpicture}}
"""

tex_template = r"""
\documentclass{{article}}
\usepackage{{geometry}}
\usepackage{{amsfonts}}
\usepackage{{amsmath}}
\usepackage{{amssymb}}
\usepackage{{sfmath}}
\usepackage{{tikz}}

% Define the set of tikz packages to be included in the architecture diagram document
\usetikzlibrary{{arrows,chains,positioning,scopes,shapes.geometric,shapes.misc,shadows}}


% Set the border around all of the architecture diagrams to be tight to the diagrams themselves
% (i.e. no longer need to tinker with page size parameters)
\usepackage[active,tightpage]{{preview}}
\PreviewEnvironment{{tikzpicture}}
\setlength{{\PreviewBorder}}{{5pt}}

\begin{{document}}

\input{{ {tikzpicture_path} }}

\end{{document}}
"""


class XDSM(object):

    def __init__(self):
        self.comps = []
        self.connections = []
        self.left_outs = {}
        self.right_outs = {}
        self.ins = {}
        self.processes = []
        self.process_arrows = []

    def add_system(
            self, node_name, style, label, stack=False, faded=False, text_width=None):
        self.comps.append([node_name, style, label, stack, faded, text_width])

    def add_input(self, name, label, style='DataIO', stack=False):
        self.ins[name] = ('output_' + name, style, label, stack)

    def add_output(
            self, name, label, style='DataIO', stack=False, side="left"):
        if side == "left":
            self.left_outs[name] = ('left_output_' + name, style, label, stack)
        elif side == "right":
            self.right_outs[name] = (
                'right_output_' +
                name,
                style,
                label,
                stack)

    def connect(
            self, src, target, label, style='DataInter', stack=False, faded=False):
        if src == target:
            raise ValueError('Can not connect component to itself')
        self.connections.append([src, target, style, label, stack, faded])

    def add_process(self, systems, arrow=True):
        self.processes.append(systems)
        self.process_arrows.append(arrow)

    def _parse_label(self, label):
        if isinstance(label, (tuple, list)):
            # mod_label = r'$\substack{'
            # mod_label += r' \\ '.join(label)
            # mod_label += r'}$'
            mod_label = r'$\begin{array}{c}'
            mod_label += r' \\ '.join(label)
            mod_label += r'\end{array}$'
        else:
            mod_label = r'${}$'.format(label)

        return mod_label

    def _build_node_grid(self):
        size = len(self.comps)

        comps_rows = np.arange(size)
        comps_cols = np.arange(size)

        if self.ins:
            size += 1
            # move all comps down one row
            comps_rows += 1

        if self.left_outs:
            size += 1
            # shift all comps to the right by one, to make room for inputs
            comps_cols += 1

        if self.right_outs:
            size += 1
            # don't need to shift anything in this case

        # build a map between comp node_names and row idx for ordering
        # calculations
        row_idx_map = {}
        col_idx_map = {}

        node_str = r'\node [{style}] ({node_name}) {{{node_label}}};'

        grid = np.empty((size, size), dtype=object)
        grid[:] = ''

        # add all the components on the diagonal
        for i_row, j_col, comp in zip(comps_rows, comps_cols, self.comps):
            style = comp[1]
            if comp[3] is True:  # stacking
                style += ',stack'
            if comp[4] is True:  # fading
                style += ',faded'
            if comp[5] is not None:
                style += ',text width={}cm'.format(comp[5])

            label = self._parse_label(comp[2])
            node = node_str.format(
                style=style,
                node_name=comp[0],
                node_label=label)
            grid[i_row, j_col] = node

            row_idx_map[comp[0]] = i_row
            col_idx_map[comp[0]] = j_col

        # add all the off diagonal nodes from components
        for src, target, style, label, stack, faded in self.connections:
            src_row = row_idx_map[src]
            target_col = col_idx_map[target]

            loc = (src_row, target_col)

            if stack is True:  # stacking
                style += ',stack'
            if faded is True:  # fading
                style += ',faded'

            label = self._parse_label(label)

            node_name = '{}-{}'.format(src, target)

            node = node_str.format(style=style,
                                   node_name=node_name,
                                   node_label=label)

            grid[loc] = node

        # add the nodes for left outputs
        for comp_name, out_data in self.left_outs.items():
            node_name, style, label, stack = out_data
            if stack:
                style += ',stack'

            i_row = row_idx_map[comp_name]
            loc = (i_row, 0)

            label = self._parse_label(label)
            node = node_str.format(style=style,
                                   node_name=node_name,
                                   node_label=label)

            grid[loc] = node

        # add the nodes for right outputs
        for comp_name, out_data in self.right_outs.items():
            node_name, style, label, stack = out_data
            if stack:
                style += ',stack'

            i_row = row_idx_map[comp_name]
            loc = (i_row, -1)
            label = self._parse_label(label)
            node = node_str.format(style=style,
                                   node_name=node_name,
                                   node_label=label)

            grid[loc] = node

        # add the inputs to the top of the grid
        for comp_name, in_data in self.ins.items():
            node_name, style, label, stack = in_data
            if stack:
                style = ',stack'

            j_col = col_idx_map[comp_name]
            loc = (0, j_col)
            label = self._parse_label(label)
            node = node_str.format(style=style,
                                   node_name=node_name,
                                   node_label=label)

            grid[loc] = node

        # mash the grid data into a string
        rows_str = ''
        for i, row in enumerate(grid):
            rows_str += "%Row {}\n".format(i) + '&\n'.join(row) + r'\\' + '\n'

        return rows_str

    def _build_edges(self):
        h_edges = []
        v_edges = []

        edge_string = "({start}) edge [DataLine] ({end})"
        for src, target, style, label, stack, faded in self.connections:
            od_node_name = '{}-{}'.format(src, target)
            h_edges.append(edge_string.format(start=src, end=od_node_name))
            v_edges.append(edge_string.format(start=od_node_name, end=target))

        for comp_name, out_data in self.left_outs.items():
            node_name, style, label, stack = out_data
            h_edges.append(edge_string.format(start=comp_name, end=node_name))

        for comp_name, out_data in self.right_outs.items():
            node_name, style, label, stack = out_data
            h_edges.append(edge_string.format(start=comp_name, end=node_name))

        for comp_name, in_data in self.ins.items():
            node_name, style, label, stack = in_data
            v_edges.append(edge_string.format(start=comp_name, end=node_name))

        paths_str = '% Horizontal edges\n' + '\n'.join(h_edges) + '\n'

        paths_str += '% Vertical edges\n' + '\n'.join(v_edges) + ';'

        return paths_str

    def _build_process_chain(self):
        sys_names = [s[0] for s in self.comps]

        chain_str = ""

        for proc, arrow in zip(self.processes, self.process_arrows):
            chain_str += "{ [start chain=process]\n \\begin{pgfonlayer}{process} \n"
            for i, sys in enumerate(proc):
                if sys not in sys_names:
                    raise ValueError(
                        'process includes a system named "{}" but no system with that name exists.'.format(sys))
                if i == 0:
                    chain_str += "\\chainin ({});\n".format(sys)
                else:
                    if arrow:
                        chain_str += "\\chainin ({}) [join=by ProcessHVA];\n".format(
                            sys)
                    else:
                        chain_str += "\\chainin ({}) [join=by ProcessHV];\n".format(
                            sys)
            chain_str += "\\end{pgfonlayer}\n}"

        return chain_str

    def write(self, file_name=None, build=True, cleanup=True, quiet=False):
        """
        Write output files for the XDSM diagram.  This produces the following:

            - {file_name}.tikz
                A file containing the TIKZ definition of the XDSM diagram.
            - {file_name}.tex
                A standalone document wrapped around an include of the TIKZ file which can
                be compiled to a pdf.
            - {file_name}.pdf
                An optional compiled version of the standalone tex file.

        Parameters
        ----------
        file_name : str
            The prefix to be used for the output files
        build : bool
            Flag that determines whether the standalone PDF of the XDSM will be compiled.
            Default is True.
        cleanup: bool
            Flag that determines if padlatex build files will be deleted after build is complete
        quiet: bool
            Set to True to suppress output from pdflatex.
        """
        nodes = self._build_node_grid()
        edges = self._build_edges()
        process = self._build_process_chain()

        module_path = os.path.dirname(__file__)
        diagram_styles_path = os.path.join(module_path, 'diagram_styles')
        # hack for Windows. MiKTeX needs Linux style paths
        diagram_styles_path = diagram_styles_path.replace('\\', '/')

        tikzpicture_str = tikzpicture_template.format(nodes=nodes,
                                                      edges=edges,
                                                      process=process,
                                                      diagram_styles_path=diagram_styles_path)

        with open(file_name + '.tikz', 'w') as f:
            f.write(tikzpicture_str)

        s_file_name = os.path.split(file_name)[-1]
        tex_str = tex_template.format(nodes=nodes, edges=edges,
                                      tikzpicture_path=s_file_name + '.tikz',
                                      diagram_styles_path=diagram_styles_path)

        if file_name:
            with open(file_name + '.tex', 'w') as f:
                f.write(tex_str)

        dirname = os.path.dirname(file_name)
        if build:
            command = 'pdflatex '
            if quiet:
                command += " -interaction=batchmode -halt-on-error "
            command += "-output-directory " + dirname + " "
            os.system(command + file_name + '.tex')
            if cleanup:
                for ext in ['aux', 'fdb_latexmk', 'fls', 'log']:
                    f_name = '{}.{}'.format(file_name, ext)
                    if os.path.exists(f_name):
                        os.remove(f_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Surrogate discipline baseclass
******************************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.core.discipline import MDODiscipline
from gemseo.core.jacobian_assembly import JacobianAssembly
from gemseo.mlearning.regression.factory import RegressionModelFactory
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()

from gemseo import LOGGER


class SurrogateDiscipline(MDODiscipline):
    """Surrogate discipline class"""

    def __init__(
        self,
        surrogate,
        data=None,
        transformer=None,
        disc_name=None,
        default_inputs=None,
        input_names=None,
        output_names=None,
        **parameters
    ):
        """
        Constructor

        :param surrogate: name of the surrogate model algorithm.
        :type surrogate: str or MLRegressionAlgo
        :param Dataset data: dataset to train the surrogate. If None,
            assumes that the surrogate is trained. Default: None.
        :param dict(str) transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :param str disc_name: name of the surrogate discipline.
            If None, use surrogate.ABBR + data.name . Default: None
        :param dict default_inputs: default inputs. If None, use the first
            sample from the dataset. Default: None.
        :param list(str) input_names: list of input names.
            If None, use all inputs. Default: None.
        :param list(str) output_names: list of output names.
            If None, use all outputs. Default: None.
        :param parameters: surrogate model parameters.
        """
        if isinstance(surrogate, MLRegressionAlgo):
            self.regression_model = surrogate
            name = self.regression_model.learning_set.name
        elif data is None:
            raise ValueError("data is required to train the surrogate model.")
        else:
            factory = RegressionModelFactory()
            self.regression_model = factory.create(
                surrogate,
                data=data,
                transformer=transformer,
                input_names=input_names,
                output_names=output_names,
                **parameters
            )
            name = self.regression_model.ABBR + "_" + data.name
        if not self.regression_model.is_trained:
            self.regression_model.learn()
            disc_name = disc_name or name
            LOGGER.info("Build the surrogate discipline: %s", disc_name)
            LOGGER.info("| Dataset name: %s", data.name)
            LOGGER.info("| Dataset size: %s", str(data.length))
            LOGGER.info(
                "| Surrogate model: %s", self.regression_model.__class__.__name__
            )
        super(SurrogateDiscipline, self).__init__(disc_name)
        self._initialize_grammars(input_names, output_names)
        LOGGER.info("| Inputs: %s", ", ".join(self.get_input_data_names()))
        LOGGER.info("| Outputs: %s", ", ".join(self.get_output_data_names()))
        self._set_default_inputs(default_inputs)
        self.add_differentiated_inputs()
        self.add_differentiated_outputs()
        try:
            self.regression_model.predict_jacobian(self.default_inputs)
            self.linearization_mode = JacobianAssembly.AUTO_MODE
            LOGGER.info("| Jacobian: surrogate model jacobian")
        except NotImplementedError:
            self.linearization_mode = self.FINITE_DIFFERENCES
            LOGGER.info("| Jacobian: finite differences")

    def __repr__(self):
        model = self.regression_model.__class__.__name__
        data_name = self.regression_model.learning_set.name
        length = len(self.regression_model.learning_set)
        msg = "SurrogateDiscipline("
        msg += "name=" + self.name + ", "
        msg += "algo=" + model + ", "
        msg += "data=" + data_name + ", "
        msg += "size=" + str(length) + ", "
        inputs = sorted(self.regression_model.input_names)
        outputs = sorted(self.regression_model.output_names)
        msg += "inputs=[" + ", ".join(inputs) + "], "
        msg += "outputs=[" + ", ".join(outputs) + "], "
        msg += "jacobian=" + self.linearization_mode
        msg += ")"
        return msg

    def __str__(self):
        data_name = self.regression_model.learning_set.name
        length = len(self.regression_model.learning_set)
        msg = "Surrogate discipline: " + self.name + "\n"
        msg += "| Dataset name: " + data_name + "\n"
        msg += "| Dataset size: " + str(length) + "\n"
        msg += "| Surrogate model: " + self.regression_model.__class__.__name__ + "\n"
        inputs = sorted(self.regression_model.input_names)
        outputs = sorted(self.regression_model.output_names)
        msg += "| Inputs: " + ", ".join(inputs) + "\n"
        msg += "| Outputs: " + ", ".join(outputs)
        return msg

    def _initialize_grammars(self, input_names=None, output_names=None):
        """ Initializes the inputs and outputs grammars from data. """
        learning_set = self.regression_model.learning_set
        in_grp = learning_set.INPUT_GROUP
        out_grp = learning_set.OUTPUT_GROUP
        if input_names is None:
            inputs = learning_set.get_data_by_group(in_grp)[0, :]
            input_names = learning_set.get_names(in_grp)
        else:
            inputs = learning_set.get_data_by_names(input_names, False)[0, :]
        if output_names is None:
            outputs = learning_set.get_data_by_group(out_grp)[0, :]
            output_names = learning_set.get_names(out_grp)
        else:
            outputs = learning_set.get_data_by_names(output_names, False)[0, :]
        inputs = DataConversion.array_to_dict(inputs, input_names, learning_set.sizes)
        outputs = DataConversion.array_to_dict(
            outputs, output_names, learning_set.sizes
        )
        self.input_grammar.initialize_from_base_dict(inputs)
        self.output_grammar.initialize_from_base_dict(outputs)

    def _set_default_inputs(self, default_inputs=None):
        """Set default inputs either from data or user specification.

        :param dict default_inputs: user default inputs.
        """
        if default_inputs is None:
            learning_set = self.regression_model.learning_set
            grp = learning_set.INPUT_GROUP
            inputs = learning_set.get_data_by_group(grp)[0, :]
            inputs = DataConversion.array_to_dict(
                inputs, learning_set.get_names(grp), learning_set.sizes
            )
            self._default_inputs = default_inputs or inputs
        else:
            self._default_inputs = default_inputs

    def _run(self):
        input_data = self.get_input_data()
        output_data = self.regression_model.predict(input_data)
        output_data = {key: val.flatten() for key, val in output_data.items()}
        self.local_data.update(output_data)

    def _compute_jacobian(self, inputs=None, outputs=None):
        input_data = self.get_input_data()
        self._init_jacobian(inputs, outputs)
        self.jac = self.regression_model.predict_jacobian(input_data)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A constraints plot
******************
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.gridspec as gridspec
import numpy as np
from future import standard_library
from matplotlib import pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.ticker import MaxNLocator

from gemseo.post.core.colormaps import PARULA, RG_SEISMIC
from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()
from gemseo import LOGGER


class ObjConstrHist(OptPostProcessor):
    """
    The **ObjConstrHist** post processing
    plots the constraint functions history in lines charts.

    By default, all constraints are considered. A sublist of constraints
    can be passed as options.
    It is possible either to save the plot, to show the plot or both.
    """

    def __init__(self, opt_problem):
        """
        Constructor

        :param opt_problem: the optimization problem to run
        """
        super(ObjConstrHist, self).__init__(opt_problem)
        self.opt_problem = opt_problem
        self.cmap = PARULA  # "viridis"  # "jet"
        self.ineq_cstr_cmap = RG_SEISMIC  # "seismic" "PRGn_r"
        self.eq_cstr_cmap = "seismic"  # "seismic" "PRGn_r"

    def _plot(
        self,
        save=False,
        show=False,
        file_path="obj_constr_hist",
        constr_names=None,
        extension="pdf",
    ):
        """
        Creates the design variables plot

        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param constr_names: names of the constraints to plot
        :type constr_names: list(str)
        :param extension: file extension
        :type extension: str
        """
        obj_name = self.opt_problem.get_objective_name()
        obj_history, x_history, n_iter = self.__get_history(obj_name)
        fmin = np.min(obj_history)
        fmax = np.max(obj_history)
        if not self.opt_problem.minimize_objective and fmax < 0.0:
            obj_history = -obj_history
            fmin = np.min(obj_history)
            fmax = np.max(obj_history)
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        fig = plt.figure(figsize=(11, 6))
        ax1 = fig.add_subplot(grid[0, 0])
        # objective function
        plt.xlabel("Iterations", fontsize=12)
        plt.ylabel("Objective value", fontsize=12)
        plt.plot(np.arange(len(obj_history)), obj_history)
        plt.ylim([fmin, fmax])
        plt.xlim([0, n_iter])
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.grid(True)
        plt.title("Evolution of the objective value and maximal constraint ")

        # Set window size
        mng = plt.get_current_fig_manager()
        mng.resize(700, 1000)

        # number of iterations
        nb_iter = x_history.shape[0]

        ineq_vals, eq_vals, ineq_cstr_id, eq_cstr_id = self.__get_constraints(
            constr_names=constr_names
        )

        # concatenate all constraints values (ineq + eq)
        # NB : we take absolute values of equality constraints for color map
        eq_vals_abs = np.abs(eq_vals)
        all_cstr_list_abs = [ineq_vals, eq_vals_abs]
        all_cstr_vals_abs = np.concatenate(
            [cstr for cstr in all_cstr_list_abs if cstr.size > 0], axis=1
        )
        maxi_by_iter = np.amax(all_cstr_vals_abs, axis=1)
        indmaxi_by_iter = np.argmax(all_cstr_vals_abs, axis=1)
        cstr_ids = np.concatenate((ineq_cstr_id, eq_cstr_id))

        values = np.atleast_2d(maxi_by_iter)
        cmap = RG_SEISMIC
        vmax = fmax
        vmin = -vmax
        extent = -0.5, nb_iter - 0.5, fmin, fmax
        im1 = ax1.imshow(
            values,
            cmap=cmap,
            interpolation="nearest",
            aspect="auto",
            extent=extent,
            norm=SymLogNorm(linthresh=1.0, vmin=vmin * 0.75, vmax=vmax * 0.75),
            alpha=0.6,
        )

        # add labels with constraint violation information
        all_cstr_list = [ineq_vals, eq_vals]
        all_cstr_vals = np.concatenate(
            [cstr for cstr in all_cstr_list if cstr.size > 0], axis=1
        )
        for iteration, ind in zip(range(nb_iter), indmaxi_by_iter):
            y_vals = 0.5 * (fmax + fmin)
            x_vals = iteration
            max_label = cstr_ids[ind]
            # no abs on equality constraint
            max_value = all_cstr_vals[iteration, ind]
            text = "constraint " + max_label + " = " + str(max_value)
            ax1.text(x_vals - 0.0, y_vals + 1.1 * fmin, text, rotation="vertical")

        # color map
        cax = fig.add_subplot(grid[0, 1])
        thick_min = int(np.log10(1.0))
        thick_max = int(np.log10(np.abs(vmax)))
        thick_num = thick_max - thick_min + 1
        levels_pos = np.logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_pos = np.append(levels_pos, vmax)
        levels_neg = np.sort(-levels_pos)
        levels_neg = np.append(levels_neg, 0)
        levels = np.concatenate((levels_neg, levels_pos))

        col_bar = fig.colorbar(im1, cax=cax, ticks=levels)
        col_bar.ax.tick_params(labelsize=9)
        cax.set_xlabel("symlog")

        # save and show
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )

    def __get_history(self, fname):
        """
        Access the optimization history of a function and the design
        variables at which it was computed

        :param fname: name of the function
        :returns: list of function values
            list of design variables values
            list of iterations
        """
        f_hist, x_hist = self.database.get_func_history(fname, x_hist=True)
        f_hist = np.array(f_hist).real
        x_hist = np.array(x_hist).real
        n_iter = x_hist.shape[0]

        return f_hist, x_hist, n_iter

    def __get_constraints(self, constr_names=None):
        """
        Returns constraints with formated shape

        :param constr_name: list of constraint names
        """
        # retrieve the constraints values
        ineq_cstr_names = []
        eq_cstr_names = []
        for cstr in self.opt_problem.get_ineq_constraints():
            if constr_names is None:
                ineq_cstr_names.append(cstr.name)
            else:
                if cstr.name in constr_names:
                    ineq_cstr_names.append(cstr.name)
        for cstr in self.opt_problem.get_eq_constraints():
            if constr_names is None:
                eq_cstr_names.append(cstr.name)
            else:
                if cstr.name in constr_names:
                    eq_cstr_names.append(cstr.name)
        get_hist_array = self.database.get_history_array
        if ineq_cstr_names != []:
            ineq_vals, ineq_id, _ = get_hist_array(ineq_cstr_names, add_dv=False)
        else:
            ineq_vals, ineq_id = np.array([]), np.array([])
        if eq_cstr_names != []:
            eq_vals, eq_cstr_id, _ = get_hist_array(eq_cstr_names, add_dv=False)
        else:
            eq_vals, eq_cstr_id = np.array([]), np.array([])

        # harmonization of tables format because constraints can be vectorial
        # or scalars. *vals.shape[0] = iteration, *vals.shape[1] = cstr values
        ineq_vals = np.atleast_3d(ineq_vals)
        ineq_shape = ineq_vals.shape
        ineq_vals = np.reshape(
            ineq_vals, (ineq_shape[0], ineq_shape[1] * ineq_shape[2])
        )
        eq_vals = np.atleast_3d(eq_vals)
        eq_shape = eq_vals.shape
        eq_vals = np.reshape(eq_vals, (eq_shape[0], eq_shape[1] * eq_shape[2]))

        return ineq_vals, eq_vals, ineq_id, eq_cstr_id



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
Polynomial regression
=====================

Polynomial regression class is a particular case of the linear regression,
where the input data is transformed before the regression is applied. This
transform consists of creating a matrix of monomials (Vandermonde) by raising
the input data to different powers up to a certain degree :math:`D`. In the
case where there is only one input variable, the input data
:math:`(x_i)_{i=1, \dots, n}\in\mathbb{R}^n` is transformed into the
Vandermonde matrix

.. math::

    \begin{pmatrix}
        x_1^1  & x_1^2  & \cdots & x_1^D\\
        x_2^1  & x_2^2  & \cdots & x_2^D\\
        \vdots & \vdots & \ddots & \vdots\\
        x_n^1  & x_n^2  & \cdots & x_n^D\\
    \end{pmatrix}
    = (x_i^d)_{i=1, \dots, n;\ d=1, \dots, D}.

The output is expressed as a weighted sum of monomials:

.. math::

     y = w_0 + w_1 x^1 + w_2 x^2 + ... + w_D x^D,

where the coefficients :math:`(w_1, w_2, ..., w_d)` and the intercept
:math:`w_0` are estimated by least square regression.

In the case of a multidimensional input, i.e.
:math:`X = (x_{ij})_{i=1,\dots,n; j=1,\dots,m}`, where :math:`n` is the number
of samples and :math:`m` is the number of input variables, the Vandermonde
matrix is expressed through different combinations of monomials of degree
:math:`d, (1 \leq d \leq D)`; e.g. for three variables
:math:`(x, y, z)` and degree :math:`D=3`, the different terms are
:math:`x`, :math:`y`, :math:`z`, :math:`x^2`, :math:`xy`, :math:`xz`,
:math:`y^2`, :math:`yz`, :math:`z^2`, :math:`x^3`, :math:`x^2y` etc. More
generally, for m input variables, the total number of monomials of degree
:math:`1 \leq d \leq D` is given by
:math:`P = \binom{m+D}{m} = \frac{(m+D)!}{m!D!}`. In the case of 3 input
variables given above, the total number of monomial combinations of degree
lesser than or equal to three is thus :math:`P = \binom{6}{3} = 20`. The linear
regression has to identify the coefficients :math:`(w_1, \dots, w_P)`, in
addition to the intercept :math:`w_0`.

This concept is implemented through the :class:`.PolynomialRegression` class
which inherits from the :class:`.MLRegressionAlgo` class.

Dependence
----------
The polynomial regression model relies on the LinearRegression class
of the `LinearRegression <https://scikit-learn.org/stable/modules/
linear_model.html>`_ and  `PolynomialFeatures <https://scikit-learn.org/stable/
modules/generated/sklearn.preprocessing.PolynomialFeatures.html>`_ classes of
the `scikit-learn library <https://scikit-learn.org/stable/modules/
linear_model.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

import pickle
from os.path import join

from future import standard_library
from numpy import concatenate, where, zeros
from sklearn.preprocessing import PolynomialFeatures

from gemseo.mlearning.regression.linreg import LinearRegression

standard_library.install_aliases()


from gemseo import LOGGER


class PolynomialRegression(LinearRegression):
    """ Polynomial regression. """

    LIBRARY = "scikit-learn"
    ABBR = "PolyReg"

    def __init__(
        self,
        data,
        degree,
        transformer=None,
        input_names=None,
        output_names=None,
        fit_intercept=True,
        penalty_level=0.0,
        l2_penalty_ratio=1.0,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param degree: Degree of polynomial. Default: 2.
        :type degree: int
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param fit_intercept: if True, fit intercept. Default: True.
        :type fit_intercept: bool
        :param penalty_level: penalty level greater or equal to 0.
            If 0, there is no penalty. Default: 0.
        :type penalty_leve: float
        :param l2_penalty_ratio: penalty ratio related to the l2
            regularization. If 1, the penalty is the Ridge penalty. If 0,
            this is the Lasso penalty. Between 0 and 1, the penalty is the
            ElasticNet penalty. Default: None.
        :type l2_penalty_ratio: float
        """
        super(PolynomialRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            fit_intercept=fit_intercept,
            penalty_level=penalty_level,
            l2_penalty_ratio=l2_penalty_ratio,
            **parameters
        )
        self.poly = PolynomialFeatures(degree=degree, include_bias=False)
        self.parameters["degree"] = degree
        if degree < 1:
            raise ValueError("Degree must be >= 1.")

    def _fit(self, input_data, output_data):
        """Fit the regression model.

        :param ndarray input_data: input data (2D).
        :param ndarray output_data: output data (2D).
        """
        input_data = self.poly.fit_transform(input_data)
        super(PolynomialRegression, self)._fit(input_data, output_data)

    def _predict(self, input_data):
        """Predict output for given input data.

        :param ndarray input_data: input data (2D).
        :return: output prediction (2D).
        :rtype: ndarray.
        """
        input_data = self.poly.transform(input_data)
        return super(PolynomialRegression, self)._predict(input_data)

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        # Dimensions:
        # powers:        (           ,            ,  n_powers ,  n_inputs )
        # coefs:         (           ,  n_outputs ,  n_powers ,           )
        # jac_coefs:     (           ,  n_outputs ,  n_powers ,  n_inputs )
        # vandermonde:   ( n_samples ,            ,  n_powers ,           )
        # contributions: ( n_samples ,  n_outputs ,  n_powers ,  n_inputs )
        # jacobians:     ( n_samples ,  n_outputs ,           ,  n_inputs )
        #
        # n_powers is given by the formula
        # n_powers = binom(n_inputs+degree, n_inputs)+1

        vandermonde = self.poly.transform(input_data)

        powers = self.poly.powers_
        n_inputs = self.poly.n_input_features_
        n_powers = self.poly.n_output_features_
        n_outputs = self.algo.coef_.shape[0]
        coefs = self.get_coefficients(False)

        jac_intercept = zeros((n_outputs, n_inputs))
        jac_coefs = zeros((n_outputs, n_powers, n_inputs))

        # Compute partial derivatives with respect to each input separately
        for index in range(n_inputs):

            # Coefficients of monomial derivatives
            dcoefs = powers[None, :, index] * coefs

            # Powers of monomial derivatives
            dpowers = powers.copy()
            dpowers[:, index] -= 1

            # Keep indices of remaining monomials only
            mask_zero = (dpowers == 0).prod(axis=1) == 1
            mask_keep = dpowers[:, index] >= 0
            mask_keep[mask_zero == 1] = False

            # Extract intercept for Jacobian (0th order term)
            dintercept = dcoefs[:, mask_zero].flatten()

            # Filter kept terms
            dcoefs = dcoefs[:, mask_keep]  # Coefficients of kept terms
            dpowers = dpowers[mask_keep]  # Power keys of kept terms

            # Find indices for the given powers
            inds_keep = [
                where((powers == dpowers[i]).prod(axis=1) == 1)
                for i in range(dpowers.shape[0])
            ]
            if len(inds_keep) > 0:
                inds_keep = concatenate(inds_keep).flatten()

            # Coefficients of partial derivatives in terms of original powers
            jac_intercept[:, index] = dintercept
            jac_coefs[:, inds_keep, index] = dcoefs

        # Assemble polynomial (sum of weighted monomials)
        contributions = jac_coefs[None] * vandermonde[:, None, :, None]
        jacobians = jac_intercept + contributions.sum(axis=2)

        return jacobians

    def get_coefficients(self, as_dict=True):
        """Return the regression coefficients of the linear fit
        as a numpy array or as a dict.

        :param bool as_dict: if True, returns coefficients as a dictionary.
            Default: True.
        """
        coefficients = self.coefficients
        if as_dict:
            raise NotImplementedError
        return coefficients

    def _save_algo(self, directory):
        """Save external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        super(PolynomialRegression, self)._save_algo(directory)
        filename = join(directory, "poly.pkl")
        with open(filename, "wb") as handle:
            pickle.dump(self.poly, handle)

    def load_algo(self, directory):
        """Load external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        super(PolynomialRegression, self).load_algo(directory)
        filename = join(directory, "poly.pkl")
        with open(filename, "rb") as handle:
            poly = pickle.load(handle)
        self.poly = poly



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
Scatter matrix
==============

The :class:`.ScatterMatrix` class implements the scatter plot matrix,
which is a way to visualize :math:`n` samples of a
multi-dimensional vector

.. math::

   x=(x_1,x_2,\ldots,x_d)\in\mathbb{R}^d

in several 2D subplots where the (i,j) subplot represents the cloud
of points

.. math::

   \left(x_i^{(k)},x_j^{(k)}\right)_{1\leq k \leq n}

while the (i,i) subplot represents the empirical distribution of the samples

.. math::

   x_i^{(1)},\ldots,x_i^{(n)}

by means of an histogram or a kernel density estimator.

A variable name can be passed to the :meth:`.DatasetPlot.execute` method
by means of the :code:`classifier` keyword in order to color the curves
according to the value of the variable name. This is useful when the data is
labeled.
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
from future import standard_library

try:
    from pandas.plotting import scatter_matrix
except ImportError:
    from pandas import scatter_matrix

from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


class ScatterMatrix(DatasetPlot):
    """ Scatter plot matrix. """

    def _plot(
        self,
        classifier=None,
        kde=False,
        size=25,
        marker="o",
        figsize_x=10,
        figsize_y=10,
    ):
        """Scatter matrix plot.

        :param classifier: variable name to build the cluster. Default: None
        :type classifier: str
        :param kde: if True, plot kernel-density estimator on the diagonal.
            Otherwise, use histograms. Default: False.
        :type kde: bool
        :param figsize_x: size of figure in horizontal direction (inches)
        :type figsize_x: int
        :param figsize_y: size of figure in vertical direction (inches)
        :type figsize_y: int
        """
        if classifier is not None and classifier not in self.dataset.variables:
            raise ValueError(
                "Classifier must be one of these names: "
                + ", ".join(self.dataset.variables)
            )
        if kde:
            diagonal = "kde"
        else:
            diagonal = "hist"
        dataframe = self.dataset.export_to_dataframe()
        if classifier is None:
            self._scatter_matrix(
                dataframe, diagonal, size, marker, figsize_x, figsize_y
            )
        else:
            self._scatter_matrix_for_group(
                classifier, dataframe, diagonal, size, marker, figsize_x, figsize_y
            )
        fig = plt.gcf()
        return fig

    def _scatter_matrix_for_group(
        self, classifier, dataframe, diagonal, size, marker, figsize_x, figsize_y
    ):
        """Scatter matrix plot for group.

        :param classifier: variable name to build the cluster. If None,
            use the first group name if the dataset is a GroupDataset,
            the first output name if the dataset is an IODataset,
            otherwise the last parameter name.
        :type classifier: str
        :param dataframe: pandas dataframe
        :param str diagonal: if 'kde', plot kernel-density estimator
            on the diagonal. If 'hist', use histograms.
        :param int figsize_x: size of figure in horizontal direction (inches)
        :param int figsize_y: size of figure in vertical direction (inches)
        """
        palette = dict(enumerate("bgrcmyk"))
        groups = self.dataset.get_data_by_names([classifier], False)[:, 0:1]
        colors = [palette[group[0] % len(palette)] for group in groups]
        _, varname = self._get_label(classifier)
        dataframe = dataframe.drop(varname, 1)
        dataframe.columns = self._get_varnames(dataframe)
        scatter_matrix(
            dataframe,
            diagonal=diagonal,
            color=colors,
            s=size,
            marker=marker,
            figsize=(figsize_x, figsize_y),
        )

    def _scatter_matrix(self, dataframe, diagonal, size, marker, figsize_x, figsize_y):
        """Scatter matrix plot for group.

        :param dataframe: pandas dataframe
        :param str diagonal: if 'kde', plot kernel-density estimator
            on the diagonal. If 'hist', use histograms.
        :param int figsize_x: size of figure in horizontal direction (inches)
        :param int figsize_y: size of figure in vertical direction (inches)
        """
        dataframe.columns = self._get_varnames(dataframe)
        scatter_matrix(
            dataframe,
            diagonal=diagonal,
            figsize=(figsize_x, figsize_y),
            s=size,
            marker=marker,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Curve plot
==========

A :class:`.Curves` plot represents samples of a functional variable
:math:`y(x)` discretized over a 1D mesh. Both evaluations of :math:`y`
and mesh are stored in a :class:`.Dataset`, :math:`y` as a parameter
and the mesh as a metadata.
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
from future import standard_library

from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


class Curves(DatasetPlot):
    """ Plot curves y_i over the mesh x. """

    def _plot(self, mesh, variable, samples=None, color="blue"):
        """Curve.

        :param mesh: name of the mesh stored in Dataset.metadata.
        :type mesh: str
        :param variable: variable name for the x-axis.
        :type variable: float
        :param samples: samples indices. If None, plot all samples.
            Default: None.
        :type samples: list(int)
        :param color: line color. Default: 'blue'.
        :type color: str
        """

        def lines_gen():
            """ Linestyle generator. """
            yield "-"
            i = 1
            for i in range(1, self.dataset.n_samples):
                yield (0, (i, 1, 1, 1))

        if samples is not None:
            output = self.dataset[variable][variable][samples, :].T
        else:
            output = self.dataset[variable][variable].T
        n_samples = output.shape[1]

        lines = lines_gen()
        for sample in range(n_samples):
            plt.plot(
                self.dataset.metadata[mesh],
                output[:, sample],
                linestyle=next(lines),
                color=color,
                label=sample,
            )
        plt.xlabel(mesh)
        plt.ylabel(variable + "(" + mesh + ")")
        plt.legend()
        fig = plt.gcf()
        return fig



# -*- coding: utf-8 -*-
"""
JSON Schema URI resolution scopes and dereferencing

https://tools.ietf.org/id/draft-zyp-json-schema-04.html#rfc.section.7

Code adapted from https://github.com/Julian/jsonschema
"""

from __future__ import absolute_import, unicode_literals, with_statement

import contextlib
import json
import re

import requests

from gemseo.utils.py23_compat import PY3

from .exceptions import JsonSchemaException

if PY3:
    import urllib.parse as urlparse
    from urllib.request import urlopen
    from urllib.parse import unquote
else:
    from urllib2 import urlopen
    import urlparse
    from urllib import unquote


def resolve_path(schema, fragment):
    """
    Return definition from path.

    Path is unescaped according https://tools.ietf.org/html/rfc6901
    """
    fragment = fragment.lstrip("/")
    parts = unquote(fragment).split("/") if fragment else []
    for part in parts:
        part = part.replace("~1", "/").replace("~0", "~")
        if isinstance(schema, list):
            schema = schema[int(part)]
        elif part in schema:
            schema = schema[part]
        else:
            raise JsonSchemaException("Unresolvable ref: {}".format(part))
    return schema


def normalize(uri):
    return urlparse.urlsplit(uri).geturl()


def resolve_remote(uri, handlers):
    """
    Resolve a remote ``uri``.

    .. note::

        Requests_ library is used to fetch ``http`` or ``https``
        requests from the remote ``uri``, if handlers does not
        define otherwise.

        For unknown schemes urlib is used with UTF-8 encoding.

    .. _Requests: http://pypi.python.org/pypi/requests/
    """
    scheme = urlparse.urlsplit(uri).scheme
    if scheme in handlers:
        result = handlers[scheme](uri)
    elif scheme in ["http", "https"]:
        result = requests.get(uri).json()
    else:
        result = json.loads(urlopen(uri).read().decode("utf-8"))
    return result


class RefResolver(object):
    """
    Resolve JSON References.
    """

    # pylint: disable=dangerous-default-value,too-many-arguments
    def __init__(self, base_uri, schema, store={}, cache=True, handlers={}):
        """
        `base_uri` is URI of the referring document from the `schema`.
        """
        self.base_uri = base_uri
        self.resolution_scope = base_uri
        self.schema = schema
        self.store = store
        self.cache = cache
        self.handlers = handlers
        self.walk(schema)

    @classmethod
    def from_schema(cls, schema, handlers={}, **kwargs):
        """
        Construct a resolver from a JSON schema object.
        """
        return cls(
            schema.get("$id", schema.get("id", "")) if isinstance(schema, dict) else "",
            schema,
            handlers=handlers,
            **kwargs
        )

    @contextlib.contextmanager
    def in_scope(self, scope):
        """
        Context manager to handle current scope.
        """
        old_scope = self.resolution_scope
        self.resolution_scope = urlparse.urljoin(old_scope, scope)
        try:
            yield
        finally:
            self.resolution_scope = old_scope

    @contextlib.contextmanager
    def resolving(self, ref):
        """
        Context manager which resolves a JSON ``ref`` and enters the
        resolution scope of this ref.
        """
        new_uri = urlparse.urljoin(self.resolution_scope, ref)
        uri, fragment = urlparse.urldefrag(new_uri)

        if normalize(uri) in self.store:
            schema = self.store[normalize(uri)]
        elif not uri or uri == self.base_uri:
            schema = self.schema
        else:
            schema = resolve_remote(uri, self.handlers)
            if self.cache:
                self.store[normalize(uri)] = schema

        old_base_uri, old_schema = self.base_uri, self.schema
        self.base_uri, self.schema = uri, schema
        try:
            with self.in_scope(uri):
                yield resolve_path(schema, fragment)
        finally:
            self.base_uri, self.schema = old_base_uri, old_schema

    def get_uri(self):
        return normalize(self.resolution_scope)

    def get_scope_name(self):
        """
        Get current scope and return it as a valid function name.
        """
        name = "validate_" + unquote(self.resolution_scope).replace("~1", "_").replace(
            "~0", "_"
        )
        name = re.sub(r"[:/#\.\-\%]", "_", name)
        name = name.lower().rstrip("_")
        return name

    def walk(self, node):
        """
        Walk thru schema and dereferencing ``id`` and ``$ref`` instances
        """
        if isinstance(node, bool):
            pass
        elif "$ref" in node and isinstance(node["$ref"], unicode):
            ref = node["$ref"]
            node["$ref"] = urlparse.urljoin(self.resolution_scope, ref)
        elif "id" in node and isinstance(node["id"], unicode):
            with self.in_scope(node["id"]):
                self.store[normalize(self.resolution_scope)] = node
                for _, item in node.items():
                    if isinstance(item, dict):
                        self.walk(item)
        else:
            for _, item in node.items():
                if isinstance(item, dict):
                    self.walk(item)
